/*
 * c_label.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_c_label.h"
#include "h_w_common.h"

#define FRM 2 /* Толщина рамки */

static XColor lbl_bg_clr; /* Цвет поля элемента */
static XColor lbl_font_clr; /* Цвет шрифта */
static XColor lbl_frm1_clr; /* Цвет рамки 1 */
static XColor lbl_frm2_clr; /* Цвет рамки 2 */
static unsigned int TXT_X = 5; /* Смещение текста относительно рамки */
static unsigned int TXT_Y = 7;


int label_get_font_height (GC gc)
{
  int ascent, descent, dir;
  win_common_get_font_info (gc, &ascent, &descent, &dir);
  return (ascent + descent + FRM*2 + MARGIN*2);
}

/* Создание текстового сообщения */
int label_create (Display * dsp, Drawable draw, int scrn, GC gc,
	     char * text, RECT * rect, LABEL * pl)
{
  if (pl->ID > 0) {
    fprintf (stderr, "%s\n", "label_create: Object already exist");
    return (1);
  }

  static unsigned long int cntID = 0;
  int depth;
  int width = rect->width;
  int height = rect->height + FRM*2 + MARGIN*2;

  int ascent, descent, dir;
  win_common_get_font_info (gc, &ascent, &descent, &dir);
  TXT_Y = MARGIN + FRM*2 + descent;
  height = rect->height = ascent + descent + FRM*2 + MARGIN*2;

  pl->rect = *rect;
  pl->txt_x = TXT_X;
  pl->txt_y = TXT_Y;
  pl->text = text;
  pl->ID = ++cntID;

  depth = DefaultDepth (dsp, scrn);
  pl->pxmFr = XCreatePixmap (dsp, draw, width, height, depth);
  pl->pxm = XCreatePixmap (dsp, draw, width-(FRM*2), height-(FRM*2), depth);

  win_common_set_color (&lbl_bg_clr, 65535, 65535, 65535);
  win_common_set_color (&lbl_font_clr, 0, 0, 0);
  win_common_set_color (&lbl_frm1_clr, 35000, 35000, 35000);
  win_common_set_color (&lbl_frm2_clr, 55000, 55000, 55000);

  label_draw (dsp, draw, gc, pl);

  return (0);
}

/* Вывод текстовой строки */
int label_draw (Display * dsp, Drawable draw, GC gc, LABEL * pl)
{
  if (pl->ID <= 0) { /* Ошибка - элемент еще не создан функцией "label_create" */
    fprintf (stderr, "%s\n", "label_draw: error - label not create");
    return (1);
  }

  int x = pl->rect.x;
  int y = pl->rect.y;
  int width = pl->rect.width;
  int height = pl->rect.height;

  /* Нарисовать рамку элемента */
  if (pl->flags & NO_FRAME) {
    XSetForeground (dsp, gc, lbl_frm2_clr.pixel);
    XFillRectangle (dsp, pl->pxmFr, gc, 0, 0, width, height);
  } else {
    XSetForeground (dsp, gc, lbl_frm1_clr.pixel);
    XFillRectangle (dsp, pl->pxmFr, gc, 0, 0, width, height);

    XSetForeground (dsp, gc, lbl_frm2_clr.pixel);
    XFillRectangle (dsp, pl->pxmFr, gc, FRM, FRM, width-FRM, height-FRM);
  }
  /* Нарисовать фон элемента */
  XSetForeground (dsp, gc, lbl_bg_clr.pixel);
  XFillRectangle (dsp, pl->pxm, gc, 0, 0, width, height);

  XSetForeground (dsp, gc, lbl_font_clr.pixel);
  XDrawString (dsp, pl->pxm, gc, pl->txt_x, height-pl->txt_y, pl->text, strlen (pl->text));

  XCopyArea (dsp, pl->pxm, pl->pxmFr, gc, 0, 0, width-(FRM*2), height-(FRM*2), FRM, FRM);
  XCopyArea (dsp, pl->pxmFr, draw, gc, 0, 0, width, height, x, y);

  return (0);
}
