/* c_tab.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

   

#include <stdio.h>
#include <malloc.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_w_main.h"
#include "h_w_common.h"
#include "h_w_tool.h"
#include "h_c_tab.h"


#define TAB_SIZE 20 /* Размер корешка вкладки (сторона квадрата). */

static int currTab; /* Текущая вкладка */


/* Создать вкладки */
int tab_create (Drawable draw, GC g, RECT t_rect, Pixmap ** px,
		unsigned long valm, XGCValues vl)
{
  int depth;
  int width = t_rect.width;
  int height = t_rect.height;
  int i; /* Индекс циклов */

  depth = DefaultDepth (display, screen);
  for (i=0; i<TAB_NUM; i++) {
    tab[i] = XCreatePixmap (display, draw, width, height, depth);
  }
  win_common_set_color (&clr, 58623, 58623, 58623);

  /* Создание функций для обработки сообщений от вкладок */
  fTAB = malloc (TAB_NUM * sizeof (void *));

  /* Установка фона для вкладок "pmx" */
  win_common_change_mask (g, valm, vl, GXcopy);
  XSetForeground (display, g, clr.pixel);
  for (i=0; i<TAB_NUM; i++) {
    XFillRectangle (display, tab[i], g, 0, 0, width, height);
  }

  currTab = 0;

  tab_init (g, px);

  return (0);
}

/* Основная функция вывода графики */
void tab_draw (Drawable draw, GC g, Pixmap * px)
{
  XCopyArea (display, *px, draw, g, 0, 0, t_win_width, t_win_height, 0, 0);
}

/* Рисование вкладок */
void tab_tabs_draw (GC g, Pixmap * px)
{
  int i, j;
  int x=TAB_SIZE, y=0, width=TAB_SIZE, height=TAB_SIZE;
  /* Нарисовать на каждой вкладке по "TAB_NUM" квадратов */
  win_common_set_color (&clr, 35535,35535,35535);
  XSetForeground (display, g, clr.pixel);
  for (i=0; i<TAB_NUM; i++) {
    px = &tab [i];
    for (j=0; j<TAB_NUM; j++)
      XDrawRectangle (display, *px, g, j*x, (t_WIN_HEIGHT-y-height), width, height);
  }
  /* Нарисовать на каждой вкладке текущий квадрат черным цветом */
  win_common_set_color (&clr, 0,0,0);
  XSetForeground (display, g, clr.pixel);
  for (i=0; i<TAB_NUM; i++) {
    px = &tab [i];
    XDrawRectangle (display, *px, g, i*x, (t_WIN_HEIGHT-y-height), width, height);
  }
}

/* Обработчик событий клавиатуры */
void tab_key_press (Pixmap ** px, KeySym keysym)
{
  if ((keysym >= XK_F1) && (keysym <= XK_F9) && ((keysym - XK_F1) < TAB_NUM)) {
    currTab = keysym - XK_F1;
    *px = &tab [currTab];
  }
}

/* Обработчик событий мыши */
void tab_button_press (GC g, Pixmap ** px, int x, int y)
{
  int coordX, coordY;

  coordX = x / TAB_SIZE; /* Номер выбранной вкладки */
  coordY = t_WIN_HEIGHT - y;

  /* Если выбрана вкладка */
  if (((coordY)/TAB_SIZE == 0) && (coordX < TAB_NUM))  {
    *px = &tab [coordX];
    currTab = coordX;
  }

  fTAB[currTab] (g, *px, x, y);
}

/* Освободить память */
void tab_mem_free (void)
{
  free (fTAB);
}
