/* c_text.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_c_text.h"
#include "h_w_main.h"
#include "h_w_common.h"

unsigned int T_TXT_X = 5; /* Смещение текста относительно рамки */
unsigned int T_TXT_Y = 7;

#define MARGIN 1 /* Полоса вокруг текста */
/* Флаги "TEXT.flags" */
#define F_CURSOR 0x0001

char tmp_str [TXT_MAX_STR];

static int fl_alt;
static XFontStruct * txtFI;
static XColor txt_bg_clr; /* Цвет поля элемента */
static XColor txt_font_clr; /* Цвет шрифта */
static unsigned long t_valmask;
static XGCValues t_val;

static void text_char_write (char ch);
static void text_back_space (void);
static void text_char_delete (void);
static void text_clear_begin (void);
static void text_clear_end (void);
static void text_clear (void);
static void text_prev_word (void);
static void text_next_word (void);

/* Создать элемент управления - текстовое поле */
int text_create (Display * dsp, Drawable draw, int scrn, GC gc,
		char * text, RECT rect, TEXT * pt)
{
  if (pt->ID > 0) {
    fprintf (stderr, "%s\n", "text_create: Object already exist");
    return (1);
  }

  static unsigned long int cntID = 0;

  int depth;
  int width = rect.width;
  int height = rect.height + MARGIN*2;

  GContext gcnt = XGContextFromGC (gc);
  txtFI =  XQueryFont (display, gcnt);
  T_TXT_Y = MARGIN + txtFI->descent;
  height = rect.height = txtFI->ascent + txtFI->descent + MARGIN*2;

  pt->rect = rect;
  pt->txt_x = T_TXT_X;
  pt->txt_y = T_TXT_Y;
  pt->text = text;
  pt->cpos = 0;
  pt->nch = strlen (text);
  pt->ID = ++cntID;

  depth = DefaultDepth (dsp, scrn);
  pt->pxm = XCreatePixmap (dsp, draw, width, height, depth);

  //win_common_set_color (&txt_bg_clr, 65535, 65535, 65535);
  win_common_set_color (&txt_bg_clr, 55535, 55535, 55535);
  win_common_set_color (&txt_font_clr, 0, 0, 0);

  /* Установка флагов */
  pt->flags = F_CURSOR;

  /* Текущий элемент управления "TEXT" (Получивший фокус). */
  pTcr = pt;

  return (0);
}

/* Вывод текстовой строки */
int text_write (Display * dsp, Drawable draw, GC gc, TEXT * pt)
{
  if (pt->ID <= 0) { /* Ошибка - элемент еще не создан функцией "label_create" */
    fprintf (stderr, "%s\n", "text_write: error - TEXT not create");
    return (1);
  }
  int cpos = pt->cpos;
  int nch = pt->nch = strlen (pt->text);
  char tch = pt->text [cpos];
  pt->text [cpos] = '\0';
  int twidth = XTextWidth (txtFI, pt->text, cpos);
  twidth = twidth + pt->txt_x;
  pt->text [cpos] = tch;

  int x = pt->rect.x;
  int y = pt->rect.y;
  int width = pt->rect.width;
  int height = pt->rect.height;

  /* Нарисовать фон элемента */
  XSetForeground (dsp, gc, txt_bg_clr.pixel);
  XFillRectangle (dsp, pt->pxm, gc, 0, 0, width, height);

  XSetForeground (dsp, gc, txt_font_clr.pixel);
  XDrawString (dsp, pt->pxm, gc, pt->txt_x, height-pt->txt_y, pt->text, nch);
  /* Нарисовать курсор ? */
  if (pt->flags & F_CURSOR) {
    win_common_change_line_width (gc, t_valmask, t_val, 1);
    XDrawLine(dsp, pt->pxm, gc, twidth, 0, twidth, height);
  }

  XCopyArea (dsp, pt->pxm, draw, gc, 0, 0, width, height, x, y);

  return (0);
}

/* Записать символ "ch" */
static void text_char_write (char ch)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);

  if (cpos > nch) cpos = nch;

  if (cpos >= 0 && (nch+1 < TXT_MAX_STR)) {
    strcpy (tmp_str, &str [cpos]);
    str [cpos] = ch;
    str [cpos+1] = '\0';
    strcat (str, tmp_str);
    cpos++;
    nch++;
  } else if (cpos >= 0 && cpos < nch) {
    str [nch-1] = '\0';
    strcpy (tmp_str, &str [cpos]);
    str [cpos] = ch;
    str [cpos+1] = '\0';
    strcat (str, tmp_str);
    cpos++;
  }

  pTcr->cpos = cpos;
}

/* Удалить символ перед курсором */
static void text_back_space (void)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);

  if (cpos > nch) cpos = nch;

  if (cpos != nch && ((cpos - 1) >= 0)) {
    str [cpos-1] = '\0';
    strcpy (tmp_str, &str [cpos]);
    strcat (str, tmp_str);
    cpos--;
    nch--;
  } else if (cpos > 0) {
    str [cpos-1] = '\0';
    cpos--;
    nch--;
  }

  pTcr->nch = nch;
  pTcr->cpos = cpos;
}

/* Удалить символ за курсором */
static void text_char_delete (void)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);

  if (cpos > nch) cpos = nch;

  if (cpos != nch && (cpos >= 0)) {
    str [cpos] = '\0';
    strcpy (tmp_str, &str [cpos + 1]);
    strcat (str, tmp_str);
  }

  pTcr->nch = nch;
  pTcr->cpos = cpos;
}

/* Очистить строку от начала и до курсора */
static void text_clear_begin (void)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);

  if (cpos > nch) cpos = nch;

  if ((cpos-1)>=0) {
    str [cpos-1] = '\0';
    strcpy (tmp_str, &str [cpos]);
    strcpy (str, tmp_str);
  }

  pTcr->nch = nch;
  pTcr->cpos = cpos = 0;
}

/* Очистить строку от курсора и до конца */
static void text_clear_end (void)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);

  if (cpos > nch) cpos = nch;

  str [cpos] = '\0';

  pTcr->nch = nch;
  pTcr->cpos = cpos;
}

/* Изменить позицию курсора */
void text_change_pos (int pos)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);

  if (pos < 0) pos = 0;
  if (pos > nch) pos = nch;

  pTcr->nch = nch;
  pTcr->cpos = cpos = pos;
}

/* Очистить строку консоли */
static void text_clear (void)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  pTcr->cpos = cpos = 0;
  str [cpos] = '\0';
  pTcr->nch = nch = strlen (str);
}

/* Перейти к предыдущему слову */
static void text_prev_word (void)
{
  int cpos = pTcr->cpos;
  char * str = pTcr->text;

  char ch;
  char chpr;
  int wfl = 0; /* Слово  */
  int bfl = 0; /* Пробел */
  int i;

  for (i = cpos-1; i >= 0; i--) {
    ch = str [i];
    if ((i-1) >= 0)
      chpr = str [i-1];
    if (ch == ' ') bfl++;
    if (bfl > 1) bfl = 0;
    if ((isalnum (ch)) && (chpr == ' ')) wfl = 1;
    if (wfl && bfl) break;
  }
  if (i < 0) i = 0;

  pTcr->cpos = cpos = i;
}


/* Перейти к следующему слову */
static void text_next_word (void)
{
  int cpos = pTcr->cpos;
  int nch =  pTcr->nch;
  char * str = pTcr->text;

  nch = strlen (str);
  char ch;
  int wfl = 0; /* Слово  */
  int bfl = 0; /* Пробел */
  int i;
#if 1
  for (i = cpos+1; i <= nch; i++) {
    ch = str [i];
    if (ch == ' ') bfl++;
    if (bfl > 1) bfl = 0;
    if ((isalnum (ch))) wfl = 1;
    if (wfl && bfl) break;
  }
  if (i > nch) i = nch;

  pTcr->cpos = cpos = i;
#endif
}

/* Обработчик событий клавиатуры */
void text_key_release (XEvent * ev, char * buff_kb, KeySym keysym)
{
  switch (keysym){
  case XK_Control_L:
    fl_control = 0;
    break;
  case XK_ISO_Prev_Group:
  case XK_Alt_L:
    fl_alt = 0;
    break;
  default:
    break;
  }
}

/* Обработчик событий клавиатуры */
void text_key_press (XEvent * ev, char * buff_kb, KeySym keysym)
{
  int cpos = pTcr->cpos;
  char * str = pTcr->text;

  switch (keysym){
  case XK_BackSpace:
    text_back_space ();
    break;
  case XK_Return:
  case 0xff8d: /* Enter на дополнительной клавиатуре */
    text_clear ();
    break;
  case XK_Home:
    text_change_pos (0);
    break;
  case XK_Up:

    break;
  case XK_Down:

    break;
  case XK_Left:
    if (fl_alt)
      text_prev_word ();
    else
      text_change_pos (--cpos);
    break;
  case XK_Right:
    if (fl_alt)
      text_next_word ();
    else
      text_change_pos (++cpos);
    break;
  case XK_End:
    text_change_pos (strlen (str));
    break;
  case XK_Escape:
    text_clear ();
    break;
  case XK_Delete:
    text_char_delete ();
    break;
  case XK_Control_L:
    fl_control = 1;
    break;
  case XK_ISO_Prev_Group:
  case XK_Alt_L:
    fl_alt = 1;
    break;
  case 0xff80 ... 0xff89:
  case 0xff91 ... 0xffaf:
  case 0xffb0 ... 0xffb9: /* Цифры на дополнительной клавиатуре */
    keysym = keysym - 0xff80;
  case 0x0020 ... 0x002f: /* Разные символы */
  case 0x0030 ... 0x0039: /* Цифры */
  case 0x003a ... 0x0040:  /* Разные символы */
  case 0x0041 ... 0x0060:   /* Буквы латиницы */
    text_char_write (keysym);
    break;
  case 0x0061: /* XK_a */
    if (fl_control)
      text_change_pos (0);
    else
      text_char_write (keysym);
    break;
  case 0x0062: /* XK_b */
    if (fl_control)
      text_change_pos (--cpos);
    else if (fl_alt)
      text_prev_word ();
    else
      text_char_write (keysym);
    break;
  case 0x0063: /* XK_c */
    text_char_write (keysym);
    break;
  case 0x0064: /* XK_d */
    if (fl_control)
      text_char_delete ();
    else
      text_char_write (keysym);
    break;
  case 0x0065: /* XK_e */
    if (fl_control)
      text_change_pos (strlen (str));
    else
      text_char_write (keysym);
    break;
  case 0x0066: /* XK_f */
    if (fl_control)
      text_change_pos (++cpos);
    else if (fl_alt)
      text_next_word ();
    else
      text_char_write (keysym);
    break;
  case 0x0067: /* XK_g */
    text_char_write (keysym);
    break;
  case 0x0068: /* XK_h */
    if (fl_control)
      text_back_space ();
    else
      text_char_write (keysym);
    break;
  case 0x0069: /* XK_i */
    text_char_write (keysym);
    break;
  case 0x006a: /* XK_j */
    if (fl_control) {
      text_clear ();
    } else
      text_char_write (keysym);
    break;
  case 0x006b: /* XK_k */
    if (fl_control)
      text_clear_end ();
    else
      text_char_write (keysym);
    break;
  case 0x006c: /* XK_l */
    text_char_write (keysym);
    break;
  case 0x006d: /* XK_m */
    text_char_write (keysym);
    break;
  case 0x006e: /* XK_n */
    text_char_write (keysym);
    break;
  case 0x006f: /* XK_o */
    text_char_write (keysym);
    break;
  case 0x0070: /* XK_p */
    text_char_write (keysym);
    break;
  case 0x0071: /* XK_q */
    text_char_write (keysym);
    break;
  case 0x0072: /* XK_r */
    text_char_write (keysym);
    break;
  case 0x0073: /* XK_s */
    text_char_write (keysym);
    break;
  case 0x0074: /* XK_t */
    text_char_write (keysym);
    break;
  case 0x0075: /* XK_u */
    if (fl_control)
      text_clear_begin ();
    else
      text_char_write (keysym);
    break;
  case 0x0076 ... 0x007a:
  case 0x007b ... 0x007e: /* Скобки */
    text_char_write (keysym);
    break;
  default:

    break;
  }

}

/* Обработчик событий мыши */
void text_button_press (int x, int y)
{
  char * str = pTcr->text;
  int i , sl, nch = strlen (str);
  int twidth;

  strcpy (tmp_str, str);

  for (i = nch; i >= 0; i--) {
    sl = strlen (tmp_str);
    twidth = XTextWidth (txtFI, tmp_str, sl);
    twidth = twidth + pTcr->txt_x;

    if (twidth > x && x < pTcr->txt_x) {
      sl = 0;
      break;
    } else if (twidth < x)
      break;

    tmp_str [i] = '\0';
  }

  text_change_pos (sl-1);
}
