/* com_console.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */
   

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_LCP_global.h"
#include "h_w_common.h"
#include "h_w_main.h"
#include "h_w_tool.h"
#include "h_w_command.h"
#include "h_LCP_reference_point.h"
#include "h_com_console.h"
#include "h_line.h"
#include "h_point.h"
#include "h_coords.h"
#include "h_fs.h"

#define PREP_IO(str1)					\
  ({if (fptr)						\
      fs_close_file (file_name, fptr);			\
    if ((fptr = fs_open_file (num_buff, str1)) != NULL)	\
      file_name = strdup (num_buff);			\
    else						\
      break;})

TABLE_PARAMETER tab_param [] = {
  {0, {"v", "view"}, "'VIEW': Enter next command"}, /* DM_VIEW = 0 */
  /* function: "ref_point_add" */    
  {1, {"p", "rpoint"}, "'POINT': Enter point's coordinate:  X   Y"}, /* DM_POINT = 1 */
  {3, {"l", "line"}, "'LINE': Enter coordinate: X1  Y1  X2  Y2"}, /* DM_LINE = 3 */
  {10, {"pl", "pline"}, "'POLYLINE': Enter next point"}, /* DM_POLYLINE = 10 */

  /* function: "ref_point_dl" */
  {101, {"dp", "rpoint_delete"}, "'DELETE POINT': Enter point's coordinate:  X   Y"}, /* DM_POINT_DELETE = 101 */
  {103, {"dl", "line_delete"}, "'DELETE LINE': Enter line's coordinate: X1 Y1 X2 Y2"}, /* DM_LINE_DELETE = 103 */
  {104, {"dld", "line_delete_dsc"}, "'DELETE LINE': Enter line's descriptor: DSC"}, /* DM_LINE_DELETE_DSC = 104 */
  {110, {"dpl", "pline_delete"}, "'DELETE POLYLINE': Enter polyline's coordinate: X1 Y1 X2 Y2"}, /* DM_POLYLINE_DELETE = 110 */
  
  {201, {"-p", "rpoint_all_delete"}, "'DELETE ALL POINTS': yes - enter '1', no - enter '0'"}, /* DM_POINT_DELETE_ALL = 201 */
  {202, {"-l", "line_all_delete"}, "'DELETE ALL LINES': yes - enter '1', no - enter '0'"}, /* DM_LINE_DELETE_ALL = 202 */
  {210, {"-pl", "pline_all_delete"}, "'DELETE ALL POLYLINES': Enter polyline's coordinate: X1 Y1 X2 Y2"}, /* DM_POLYLINE_DELETE_ALL = 210 */
  
  {310, {"st", "show_tool"}, "'SHOW TOOL WINDOW': OK"}, /* DM_SHOW_W_TOOL = 310 */
  {311, {"ut", "unshow_tool"}, "'UNSHOW TOOL WINDOW': OK"}, /* DM_UNSHOW_W_TOOL = 311 */

  {401, {"of", "fs_open_file"}, "'OPEN FILE': OK"}, /* FS_READ_FILE = 401 */
  {402, {"cf", "fs_close_file"}, "'CLOSE FILE': OK"}, /* FS_CLOSE_FILE = 402 */
  {403, {"wf", "fs_write_file"}, "'WRITE FILE': OK"}, /* FS_WRITE_FILE = 403 */

  /* Объект точка применяется для создания объектов:
     линия, полилиния, прямоугольник, дуга, окружность, кривая.
     Непосредственно на экране не отображается.
     function: "point_add" */    
  {500, {"pt", "point"}, "'TEST POINT ADD':  Enter point's coordinate:  X   Y"}, /* DM_TEST_POINT = 500 */

  {550, {"dt", "point_delete"}, "'TEST POINT DELETE': Enter point's descriptor: DSC"}, /* DM_TEST_POINT_DELETE = 550 */
  
  {901, {"99", "--nowindow"}, ""}, /* COM_NO_WINDOW = 901 */
  {902, {"98", "--comcon"}, ""}, /* COM_WND_CON = 902 */
  {903, {"97", "--select-font"}, ""}, /* COM_SELECT_FONT = 903 */  
  {904, {"96", "--pan-draw"}, ""}, /* COM_PAN_DRAW = 904 */  
  
  {950, {"50", "--version"}, LCP_VERSION}, /* COM_VERSION = 950 */

  /* function: "line_test_print_certain" */
  {994, {"6", "printline"}, "'PRINT LINE': OK"}, /* DM_LINE_TEST_PRINT_CERTAIN = 994 */
  /* function: "line_test_print_all" */
  {995, {"5", "printlines"}, "'PRINT LINES': OK"}, /* DM_PRINT_LINES = 995 */
  /* function: "line_test_print_all" */
  {996, {"4", "PrintLines"}, "'TEST PRINT LINES': OK"}, /* DM_TEST_PRINT_LINES = 996 */
  /* function: "point_print_test" (Тестирование списка точек) */
  {997, {"3", "point_print_test"}, "'TEST PRINT POINTS': OK"}, /* DM_TEST_PRPOINTS = 997 */
  /* function: "ref_point_print2List" */
  {998, {"2", "pr2points"}, "'POINT PRINT2': OK"}, /* DM_POINT_PRINT2 = 998 */
  /* function: "ref_point_printList" */
  {999, {"1", "prpoints"}, "'POINT PRINT': OK"}, /* DM_POINT_PRINT = 999 */
  {1000, {"q", "quit"}, "'QUIT': yes - enter '1', no - enter '0'"}, /* DM_QUIT = 1000 */
  {-1, {"0", "0"}, ""},
};


static void com_con_get_command (void);
static void com_con_parse_command (void);
static void com_con_get_str (void);
static int com_con_get_nums (int cnt);


/* Работа с консолью (чтение и запись значений). */
int com_con(void)
{
  while (1) {
    com_con_print ("ENTER NEXT COMMAND");
    com_con_get_command ();
#if _DEBUG
    fprintf (stderr, "%s %d\n", command_buff, strlen(command_buff));
    fflush (stderr);
#endif
    com_con_parse_command ();
  }

  return (0);
}

/* Получение команды из консоли.
   Если ch имеет тип int, то
   в данном случае необходимо выравнивание данных путем
   введения дополнительного параметра типа int.
   Если количество введенных символов превышает размер
   буфера command_buff, то лишние символы удаляются. */
static void com_con_get_command (void)
{
  unsigned char ch;
  int i;
  int STRNG;

  STRNG = BUFF_SIZE - 1;

  i = 0;
  while ((read (0, &ch, 1)) > 0 && ch != '\n') {
    command_buff [i] = ch;
    i++;
    if (i >= STRNG) {
      while ((read (0, &ch, 1)) > 0 && ch != '\n')
	;
      break;
    }
  }

  command_buff [i] = '\0';
}


/* Распознание вводимых интерактивных
   параметров. */
static void com_con_parse_command (void)
{
  int i;

  if (strlen (command_buff) <= SHORT_PAR_LEN) {
    i = com_con_search_param (0, command_buff); /* поиск в списке коротких параметров */
#if _DEBUG
    fprintf (stderr, "%d\n", i);
    fflush (stderr);
#endif
  } else {
    i = com_con_search_param (1, command_buff); /* поиск в списке длинных параметров */
#if _DEBUG
    fprintf (stderr, "%d\n", i);
    fflush (stderr);
#endif
  }
  if (tab_param [i].num == (-1))
    return;
  com_con_do_list (i);
}

/* Поиск параметра в массиве. */
int com_con_search_param (int flg, char * bf)
{
  int i;
  int n;

  i = 0;
  while (tab_param [i].param [flg] [0] != '0') {
    n = strcmp (tab_param [i].param [flg], bf);
    if (n == 0)
      return (i);
    i++;
  }
  return (i);
}

/*	Вызов обработчика соответствующего
  	команде. */
void com_con_do_list (int i)
{
    if (XComCon || XWndComCon) {
      com_con_print (tab_param [i].hlpmsg);
    }
      
  switch (tab_param [i].num) {
  case DM_VIEW : /* Режим просмотра. */
    break;
  case DM_POINT : /* Нарисовать точку. "reference_point.c" */
    if (XComCon) {
      if ((com_con_get_nums (2)) == RET_ERROR) return; /* Получить два числа. */
    } else if (XWndComCon) {
      if ((com_con_get_nums (2)) == RET_ERROR) return; /* Получить два числа. */
      coords_draw_point_world (nums.ord [0], nums.ord [1]);
      win_common_redraw_window (m_window);
    } else if (XWndWork) {
      if ((win_com_get_nums (2)) == CON_RET_ERROR) return;
      coords_draw_point_world (nums.ord [0], nums.ord [1]);
    }
    ref_point_add (nums.ord [0], nums.ord [1]);  /* Занести данные в структуры. */
    break;
  case DM_LINE : /* Нарисовать линию. */
    if (XComCon) {
      if ((com_con_get_nums (4)) == RET_ERROR) return; /* Получить четыре числа. */
    } else if (XWndComCon) {
      if ((com_con_get_nums (4)) == RET_ERROR) return; /* Получить четыре числа. */
    } else if (XWndWork) {
      if ((win_com_get_nums (4)) == CON_RET_ERROR) return;
    }
    //   printf ("%f %f %f %f \n", nums.ord [0], nums.ord [1], nums.ord [2], nums.ord [3]);
    line_draw (&lLine, tPixval, nums.ord [0], nums.ord [1], nums.ord [2], nums.ord [3]);
    if (XWndComCon || XWndWork)
      coords_draw_line4_world (nums.ord [0], nums.ord [1], nums.ord [2], nums.ord [3]);
    break;
  case DM_POLYLINE : /* Нарисовать полилинию. */
    break;

  case DM_POINT_DELETE : /* Удалить точку с указанными координатами. */
    if (XComCon) {
      if ((com_con_get_nums (2)) == RET_ERROR) return;
      ref_point_dl (nums.ord [0], nums.ord [1]);
    } else if (XWndComCon)  {
      if ((com_con_get_nums (2)) == RET_ERROR) return;
      ref_point_dl (nums.ord [0], nums.ord [1]);
      coords_draw ();
      win_common_redraw_window (m_window);
    } else if (XWndWork) {
      if ((win_com_get_nums (2)) == CON_RET_ERROR) return;
      ref_point_dl (nums.ord [0], nums.ord [1]);
      coords_draw ();      
    }
    break;
  case DM_LINE_DELETE: /* Удалить линию с координатами */
    if (XComCon) {
      if ((com_con_get_nums (4)) == RET_ERROR) return;
    } else if (XWndComCon) {
      if ((com_con_get_nums (4)) == RET_ERROR) return;
    } else if (XWndWork) {
      if ((win_com_get_nums (4)) == CON_RET_ERROR) return;
    }
    line_delete4 (&lLine, nums.ord [0], nums.ord [1], nums.ord [2], nums.ord [3]);
    if (XWndComCon || XWndWork)
      coords_draw ();
    break;
  case DM_LINE_DELETE_DSC: /* Удалить линию с дескриптором */
    if (XComCon) {
      if ((com_con_get_nums (1)) == RET_ERROR) return;
    } else if (XWndComCon) {
      if ((com_con_get_nums (1)) == RET_ERROR) return;      
    } else if (XWndWork) {
      if ((win_com_get_nums (1)) == CON_RET_ERROR) return;
    }
    line_delete (&lLine, (unsigned long) nums.ord [0]);
    if (XWndComCon || XWndWork)
      coords_draw ();
    break;
  case DM_POLYLINE_DELETE: /* Удалить полилинию */
    
    break;
  case DM_POINT_DELETE_ALL : /* Удалить все точки */
    if (XComCon) {
      ref_point_del_all ();
    } else if (XWndComCon) {
      ref_point_del_all ();
      coords_delete_all_points ();
      win_common_redraw_window (m_window);
    } else if (XWndWork) {
      if ((win_com_get_nums (1)) == CON_RET_ERROR) return;
      if (nums.ord [0] == 1) {
	ref_point_del_all ();
	coords_delete_all_points ();
      }
    }
    break;
  case DM_LINE_DELETE_ALL: /* Удалить все линии */
    if (XComCon) {
      if ((com_con_get_nums (1)) == RET_ERROR) return;
    } else if (XWndComCon) {
      if ((com_con_get_nums (1)) == RET_ERROR) return;
    } else if (XWndWork) {
      if ((win_com_get_nums (1)) == CON_RET_ERROR) return;
    }
    if ((win_com_get_nums (1)) == CON_RET_ERROR) return;
    if (nums.ord [0] == 1) {
      line_delete_all (&lLine);
      if (XWndComCon || XWndWork)
	coords_draw ();
    }
    break;
  case DM_POLYLINE_DELETE_ALL: /* Удалить все полилинии */
    
    break;
  case DM_SHOW_W_TOOL: /* Показать окно с инструментами. */
    if (XWndWork)
      win_tool_show ();
    break;
  case DM_UNSHOW_W_TOOL: /* Убрать окно с инструментами */
    if (XWndWork)
      win_tool_unshow ();
    break;
  case FS_READ_FILE: /* Открыть файл (существующий файл предварительно закрывается) */
    PREP_IO ("r");
    fs_read_file (fptr);
    coords_draw ();    
    break;
  case FS_CLOSE_FILE: /* Закрыть файл */
    fs_close_file (file_name, fptr);
    coords_draw ();    
    break;
  case FS_WRITE_FILE: /* Запись в файл */
    PREP_IO ("w");
    fs_write_file (fptr);
    break;    

  case DM_TEST_POINT : /* Ввести координаты тестируемой точки */
    if (XComCon) {
      if ((com_con_get_nums (2)) == RET_ERROR) return;
    } else if (XWndComCon) {
      if ((com_con_get_nums (2)) == RET_ERROR) return;
    } else if (XWndWork) {
      if ((win_com_get_nums (2)) == CON_RET_ERROR) return;
    }
    point_add (&tstP, nums.ord [0], nums.ord [1]);

    break;
  case DM_TEST_POINT_DELETE : /* Удаление точки с соответствующим дескриптором из списка */
    if (XComCon) {
      if ((com_con_get_nums (1)) == RET_ERROR) return;
    } else if (XWndComCon) {
      if ((com_con_get_nums (1)) == RET_ERROR) return;
    } else if (XWndWork) {
      if ((win_com_get_nums (1)) == CON_RET_ERROR) return;
    }
    printf ("%f \n", nums.ord [0]);
    fflush (stdout);
    tstP = point_delete (&tstP, nums.ord [0]);
    break;
  case DM_LINE_TEST_PRINT_CERTAIN: /* Распечатать параметры указанной линии */
    if (XComCon) {
      if ((com_con_get_nums (4)) == RET_ERROR) return;
    } else if (XWndComCon) {
      if ((com_con_get_nums (4)) == RET_ERROR) return;
    } else if (XWndWork) {
      if ((win_com_get_nums (4)) == CON_RET_ERROR) return;
    }
    line_test_print_certain (&lLine, nums.ord [0], nums.ord [1], nums.ord [2], nums.ord [3]);
    break;
  case DM_PRINT_LINES: /* Распечатать список линий */
    //printf ("PRINT LINES: %d\n", DM_PRINT_LINES);
    line_test_print_all (&lLine);
    break;
  case DM_TEST_PRINT_LINES: /* Распечатать список линий */
    line_test_print_all (&lLine);
    break;
  case DM_TEST_PRPOINTS:
    point_print_test (&tstP);
    break;
  case DM_POINT_PRINT2 : /* Проверить указатели "prev" и "next" в списке */
    ref_point_print2List ();
    break;
  case DM_POINT_PRINT : /* Распечатать список точек */
    ref_point_printList ();
    break;
  case DM_QUIT : /* Выход. */
    if (XWndComCon || XComCon) {
      exit(0);
    } else if (XWndWork) {
      if ((win_com_get_nums (1)) == CON_RET_ERROR) return;
      if (nums.ord [0] == 1) {
	XCloseDisplay(display);
	exit(0);
      }
    }
    break;
  default:
    com_con_err_print ("BAD COMMAND");
  }
}

/* Функция работает аналогично функции "com_con_get_command".  См. комментарий
   "com_con_get_command". */
static void com_con_get_str (void)
{
  unsigned char ch;
  int i;
  int STRNG;


  STRNG = NUMBUFFSIZE - 1;

  i = 0;
  while ((read (0, &ch, 1)) > 0 && ch != '\n') {
    num_buff [i] = ch;
    i++;
    if (i >= STRNG) {
      while ((read (0, &ch, 1)) > 0 && ch != '\n')
	;
      break;
    }
  }

  num_buff [i] = '\0';
}

/* Функция получает несколько чисел,
   проверяет корректность ввода, допустимые пределы.
   В случае ввода неверных данных функция выводит
   сообщение об ошибке. */
static int com_con_get_nums (int cnt)
{
  int state = OUT;
  int i = 0;
  char ch;
  char lbuff [LNUM_MAX];
  int lnum = 0;
  int order = 0;

  com_con_get_str ();

  if (cnt >= 1 && cnt < 5)
    nums.cnt = cnt;
  else {
    com_con_err_print ("error: parameters incorrect");
    return (-1);
  }

  while ((ch = num_buff [i]) != '\0' ) {
    if (isdigit (ch) || ch == '-' || ch == '.' || ch == ',' ) {
      if (ch == ',') ch = '.';
      if (state == OUT) {
	switch (order) {
	case 0:
	  lnum = 0;
	  order++;
	  break;
	case 1:
	case 2:
	case 3:
	case 4:
	  lbuff [lnum] = '\0';
	  nums.ord [order-1] = atof (lbuff);
	  lnum = 0;
	  order++;
	  break;
	default:
	  com_con_err_print ("error: too many parameters");
	  return (-1);
	}
      }
      lbuff [lnum] = ch;
      lnum++;
      if (lnum > LNUM_MAX-1) {
	com_con_err_print ("error: too long number");
	return (-1);
      }
      state = IN;
    } else {
      state = OUT;
    }

    i++;
  } // while


  lbuff [lnum] = '\0';
  nums.ord [order-1] = atof (lbuff);

  if (order < 1) {
    com_con_err_print ("error: too little parameters");
    return (-1);
  }

  return (0);
}

/* Удалить пробелы во всей строке */
char * com_con_no_blank (char * sstr)
{
  int i = 0, j = 0;
  char ch;

  while (1) {
    ch = sstr [i];
    if (ch == '\0') break;
    if (ch != ' ') {
      sstr [j] = sstr [i];
      j++;
    }
    i++;
  }
  sstr [j] = '\0';

  return (sstr);
}

/* Вывод сообщения. */
void com_con_print (char str [])
{
  fprintf (stdout, "%s\n", str);
  fflush (stdout);
}

/* Сообщение об ошибке. */
void com_con_err_print (char str [])
{
  fprintf (stderr, "%s\n", str);
  fflush (stderr);
}
