/* coords.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

/*   
   Вспомогательные функции главного окна "w_main.c".
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>

#include "h_w_main.h"
#include "h_w_tool.h"
#include "h_w_command.h"
#include "h_w_common.h"
#include "h_LCP_reference_point.h"
#include "h_o_cursor.h"
#include "h_LCP_f_geometry.h"
#include "h_coords.h"
#include "h_line.h"

#define wshrink 0
#define xmin wshrink
#define xmax (m_win_width - wshrink)
#define ymin wshrink
#define ymax (m_win_height - wshrink)

const float hSz = 2.00000;  /* Размер точки */
const float crSz = 15.00000;  /* Размер точки отсчета */
static unsigned long __red, __green, __blue;

static void coords_draw_line (pLINE l);
static void coords_draw_point (float X, float Y);
static void coords_origin_place (float X, float Y);
static int  coords_clip_get_code (float x, float y);
static RECT_POINTS coords_clip (int * flg, float x1, float y1, float x2, float y2);
static void coords_draw_line4 (float X1, float Y1, float X2, float Y2);

/* Заполнение структуры координат */
void coords_init (void)
{
  m_scale = 1;
  scl = 1;

  zeroPt.x = 0;
  zeroPt.y = 0;
}

/* Panorama */
void coords_pan_draw_lines (void)
{
  win_main_set_background ();
  CrCurs.flag = 0;
  register pRefPointList ptr;
  register pLINE pln;
  win_main_change_mask (GXcopy);
  XSetForeground (display, m_gc, pt_clr.pixel);
  scl = 1;

  /* Пересчет координат */  
  float CrCurs_X = ((float)CrCurs.X - (float)CrCurs.oldX);
  float CrCurs_Y = -((float)CrCurs.Y - (float)CrCurs.oldY);
  
  /* Начало координат */
  zeroPt.x = CrCurs_X + zeroPt.x;
  zeroPt.y = CrCurs_Y + zeroPt.y;
  coords_origin_place (zeroPt.x, zeroPt.y);

  /* Точки для ориентирования на плоскости */
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
    ptr->pp.x = CrCurs_X + ptr->pp.x;
    ptr->pp.y = CrCurs_Y + ptr->pp.y;
    coords_draw_point (ptr->pp.x, ptr->pp.y);
  }

  /* Линии */
  for (pln = lLine; pln != NULL; pln = pln->next) {
    pln->p1->pp.x = CrCurs_X + pln->p1->pp.x;
    pln->p1->pp.y = CrCurs_Y + pln->p1->pp.y;
    pln->p2->pp.x = CrCurs_X + pln->p2->pp.x;
    pln->p2->pp.y = CrCurs_Y + pln->p2->pp.y;
    coords_draw_line (pln);
  }

}

/* Panorama after
   Функция перерисовывает экран после перемещения (Функция "Panorama"). */
void coords_pan_draw_lines_after (void)
{
  
}

/* Panorama */
void coords_pan_no_lines (void)
{
  win_main_set_background ();
  CrCurs.flag = 0;
  register pRefPointList ptr;
  register pLINE pln;
  win_main_change_mask (GXcopy);
  XSetForeground (display, m_gc, pt_clr.pixel);
  scl = 1;

  /* Пересчет координат */  
  float CrCurs_X = ((float)CrCurs.X - (float)CrCurs.oldX);
  float CrCurs_Y = -((float)CrCurs.Y - (float)CrCurs.oldY);
  
  /* Начало координат */
  zeroPt.x = CrCurs_X + zeroPt.x;
  zeroPt.y = CrCurs_Y + zeroPt.y;

  /* Точки для ориентирования на плоскости */
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
    ptr->pp.x = CrCurs_X + ptr->pp.x;
    ptr->pp.y = CrCurs_Y + ptr->pp.y;
  }

  /* Линии */
  for (pln = lLine; pln != NULL; pln = pln->next) {
    pln->p1->pp.x = CrCurs_X + pln->p1->pp.x;
    pln->p1->pp.y = CrCurs_Y + pln->p1->pp.y;
    pln->p2->pp.x = CrCurs_X + pln->p2->pp.x;
    pln->p2->pp.y = CrCurs_Y + pln->p2->pp.y;
  }
}

/* Panorama after
   Функция перерисовывает экран после перемещения (Функция "Panorama"). */
void coords_pan_no_lines_after (void)
{
  win_main_set_background ();
  CrCurs.flag = 0;
  register pRefPointList ptr;
  register pLINE pln;
  win_main_change_mask (GXcopy);
  XSetForeground (display, m_gc, pt_clr.pixel);
  scl = 1;

  /* Пересчет координат */  
  coords_origin_place (zeroPt.x, zeroPt.y);

  /* Точки для ориентирования на плоскости */
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
    coords_draw_point (ptr->pp.x, ptr->pp.y);
  }

  /* Линии */
  for (pln = lLine; pln != NULL; pln = pln->next) {
    coords_draw_line (pln);
  }
}

/* Zoom
   Функция масштабирования экранного изображения. */
void coords_zoom (void)
{
  win_main_set_background ();
  CrCurs.flag = 0;
  register pRefPointList ptr;
  register pLINE pln;
  win_main_change_mask (GXcopy);
  XSetForeground (display, m_gc, pt_clr.pixel);
  RECT_POINTS r;
  int flg;
  float rCrX_ch, rCrY_ch;
  
  /* Начало координат */
  zeroPt.x = (float)rCrX - (((float)rCrX - zeroPt.x) * scl);
  zeroPt.y = (float)rCrY - (((float)rCrY - zeroPt.y) * scl);
  coords_origin_place (zeroPt.x, zeroPt.y);

  /* Пересчет координат */
  rCrX_ch = (float)rCrX - ((float)rCrX * scl);
  rCrY_ch = (float)rCrY - ((float)rCrY * scl);

  /* Точки для ориентирования на плоскости */
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
    ptr->pp.x = rCrX_ch + ptr->pp.x * scl;
    ptr->pp.y = rCrY_ch + ptr->pp.y * scl;
    coords_draw_point (ptr->pp.x, ptr->pp.y);
  }

  /* Линии */
  for (pln = lLine; pln != NULL; pln = pln->next) {
    pln->p1->pp.x = rCrX_ch + pln->p1->pp.x * scl;
    pln->p1->pp.y = rCrY_ch + pln->p1->pp.y * scl;
    pln->p2->pp.x = rCrX_ch + pln->p2->pp.x * scl;
    pln->p2->pp.y = rCrY_ch + pln->p2->pp.y * scl;

    r = coords_clip (&flg, pln->p1->pp.x, pln->p1->pp.y, pln->p2->pp.x, pln->p2->pp.y);
    if (flg == 0) {
      __red = pln->r; __green = pln->g; __blue = pln->b;
      coords_draw_line4 (r.x1, r.y1, r.x2, r.y2);
    }
  }
  win_main_draw ();

  /* Проверка наличия графической консоли, установка параметров в консоль. */ 
  if (!XWndComCon) win_com_mode_def (); 
}

/* Zoom In
   Увеличить изображение на экране */
void coords_zoom_in (void)
{
  scl = 1 * m_scale_coef; /* printf ("%f  ", scl); fflush(stdout); */
  m_scale *= m_scale_coef;

  coords_zoom ();
}

/* Zoom Out
   Уменьшить изображение на экране */
void coords_zoom_out (void)
{
  if (m_scale < 0.0001) return;  
  
  scl = 1 / m_scale_coef; /* printf ("%f  ", scl); fflush(stdout); */
  m_scale /= m_scale_coef; 
  
  coords_zoom ();
}

/* Получение информации о положении точки относительно окна */
static int coords_clip_get_code (float x, float y)
{
  return (x<xmin) << 3 | (x>xmax) << 2 | (y<ymin) << 1 | (y>ymax);
}
/* Анализ заданного отрезка прямой линии для вычерчивания только
   той его части, которая заключена внутри окна */
static RECT_POINTS
coords_clip (int * flg, float x1, float y1, float x2, float y2)
{
  int c1 = coords_clip_get_code (x1, y1), c2 = coords_clip_get_code (x2, y2);
  float dx, dy;
  RECT_POINTS r;
  *flg = 0;
  
  while (c1 | c2) {
    if (c1 & c2)
      {
	*flg = -1;
	goto report;
      }
    dx = x2 - x1; dy = y2 - y1;
    if (c1) {
      if (x1 < xmin) { y1 += dy * (xmin - x1) / dx; x1 = xmin; }
      else if (x1 > xmax) { y1 += dy * (xmax - x1) / dx; x1 = xmax; }
      else if (y1 < ymin) { x1 += dx * (ymin - y1) / dy; y1 = ymin; }
      else if (y1 > ymax) { x1 += dx * (ymax - y1) / dy; y1 = ymax; }
      c1 = coords_clip_get_code (x1, y1);
    } else {
      if (x2 < xmin) { y2 += dy * (xmin - x2) / dx; x2 = xmin; }
      else if (x2 > xmax) { y2 += dy * (xmax - x2) / dx; x2 = xmax; }
      else if (y2 < ymin) { x2 += dx * (ymin - y2) / dy; y2 = ymin; }
      else if (y2 > ymax) { x2 += dx * (ymax - y2) / dy; y2 = ymax; }
      c2 = coords_clip_get_code (x2, y2);
    }
  }
  r.x1 = x1; r.y1 = y1; r.x2 = x2; r.y2 = y2;

 report:
  return (r);
}

/* Отобразить знак начала координат */
static void coords_origin_place (float X, float Y)
{
  win_main_line ((int)(X), (int)(Y-crSz), (int)(X), (int)(Y+crSz));
  win_main_line ((int)(X-crSz), (int)(Y), (int)(X+crSz), (int)(Y));
}

/* Нарисовать точку */
static void coords_draw_point (float X, float Y)
{
    win_main_line ((int)(X), (int)(Y-hSz), (int)(X), (int)(Y+hSz));
    win_main_line ((int)(X-hSz), (int)(Y), (int)(X+hSz), (int)(Y));
}

/* Нарисовать точку (координаты мировые) */
void coords_draw_point_world (float X, float Y)
{
  win_main_change_mask (GXcopy);
  XSetForeground (display, m_gc, pt_clr.pixel);

  X = X * m_scale + zeroPt.x; Y = Y * m_scale + zeroPt.y; /* Пересчет из мировых координат в экранные */   

  win_main_line ((int)(X), (int)(Y-hSz), (int)(X), (int)(Y+hSz));  /* Горизонтальная линия */ 
  win_main_line ((int)(X-hSz), (int)(Y), (int)(X+hSz), (int)(Y));  /* Вертикальная линия */
  win_main_draw ();
}

/* Отобразить все точки */
void coords_draw (void)
{
  win_main_set_background ();
  CrCurs.flag = 0;

  register pRefPointList ptr;
  XSetForeground (display, m_gc, pt_clr.pixel);
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
    win_main_line ((int)(ptr->pp.x), (int)(ptr->pp.y-hSz), (int)(ptr->pp.x), (int)(ptr->pp.y+hSz));
    win_main_line ((int)(ptr->pp.x-hSz), (int)(ptr->pp.y), (int)(ptr->pp.x+hSz), (int)(ptr->pp.y));
  }

  register pLINE pln;
  for (pln = lLine; pln != NULL; pln = pln->next) {
    coords_draw_line (pln);
  }
  
  coords_origin_place (zeroPt.x, zeroPt.y);

  win_main_draw ();
}

/* Нарисовать линию */
static void coords_draw_line (pLINE l)
{
  XColor c;
  win_common_set_color (&c, l->r, l->g, l->b); XSetForeground (display, m_gc, c.pixel);
  win_main_line ((int)(l->p1->pp.x), (int)(l->p1->pp.y), (int)(l->p2->pp.x), (int)(l->p2->pp.y));
  win_common_set_color (&c, 65535, 65535, 65535); XSetForeground (display, m_gc, c.pixel);
}

/* Нарисовать линию (Функция подобная coords_draw_line, с четырмя параметрами) */
static void coords_draw_line4 (float X1, float Y1, float X2, float Y2)
{
  XColor c;
  win_common_set_color (&c, __red, __green, __blue); XSetForeground (display, m_gc, c.pixel);
  win_main_line ((int)(X1), (int)(Y1), (int)(X2), (int)(Y2));
  win_common_set_color (&c, 65535, 65535, 65535); XSetForeground (display, m_gc, c.pixel);
  win_main_draw ();
}

/* Нарисовать линию (Функция подобная coords_draw_line4 (координаты мировые)) */
void coords_draw_line4_world (float X1, float Y1, float X2, float Y2)
{
  XColor c;
  X1 = X1 * m_scale + zeroPt.x;
  X2 = X2 * m_scale + zeroPt.x;
  Y1 = Y1 * m_scale + zeroPt.y;
  Y2 = Y2 * m_scale + zeroPt.y;

  win_common_set_color (&c, red, green, blue); XSetForeground (display, m_gc, c.pixel);
  win_main_line ((int)(X1), (int)(Y1), (int)(X2), (int)(Y2));
  win_common_set_color (&c, 65535, 65535, 65535); XSetForeground (display, m_gc, c.pixel);
  win_main_draw ();
}

/* Удалить все точки */
void coords_delete_all_points (void)
{
  win_main_set_background ();
  CrCurs.flag = 0;
  win_main_draw ();
}
