/* fs.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include "h_w_common.h"
#include "h_line.h"
#include "h_LCP_global.h"
#include "h_LCP_reference_point.h"
#include "h_fs.h"
#include "h_coords.h"


#define TRACE_PRINT 0
#define MESSAGE_FS(str) ({fprintf (stdout, "%s\n", str); fflush (stdout);})
#define ERR_FS(str, fn, line, en)					\
  ({fprintf (stderr, "ERR: %s %s (source file: %s, line: %d, errno: %d) \n", \
	    str, fn, __FILE__, line, en);				\
    fflush (stderr); })

static void fs_parse (const char * buff);
static int fs_next_pos (int i, const char * buff);
static int fs_scroll_str (int i, const char * buff);

/* Открыть файл с именем "fn" */
FILE * fs_open_file (char * fn, char * mode)
{
  FILE * f;
# if TRACE_PRINT
  char bf [TXT_MAX_STR];
#endif

  f = fopen (fn, mode);
  if (f == NULL) {
    ERR_FS ("can't open file", fn, __LINE__, errno);
    goto ret;
  }

# if TRACE_PRINT
  while (fgets (bf, TXT_MAX_STR-1, f)) {
    fs_parse (bf);
    printf ("%s \n", bf);
  }
# endif

 ret:
  return (f);
}

/* Закрыть файл */
int fs_close_file (char * fn, FILE * fp)
{
  int retval = 0;
  if (fp == NULL) { /* Значение "errno" неопределенное */
    ERR_FS ("file pointer is NULL, can't close file ", fn, __LINE__, errno);
    return (-1);
  }
  else {
    retval = fclose (fp);
    if (retval == EOF)
      ERR_FS ("can't close file", fn, __LINE__, errno);
    free (file_name);
    file_name = (char *) NULL;
    MESSAGE_FS ("FILE CLOSED");
  }

  ref_point_del_all ();
  line_delete_all (&lLine);
  
  return (retval);
}

/* Чтение файла */
int fs_read_file (FILE * fp)
{
  char bf [TXT_MAX_STR];

  coords_init ();
  
  while (fgets (bf, TXT_MAX_STR-1, fp)) {
    fs_parse (bf);
    printf ("%s \n", bf);
  }
  
  return (0);
}

/* Запись в файл */
int fs_write_file (FILE * fp)
{
  register pRefPointList ptr;
  register pLINE pln;

  /* Точки для ориентирования на плоскости */
  if (refPoint) {
    for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
      fprintf (fp, "%s %f %f \n",
	       "POINT", ptr->pt.x, ptr->pt.y);
    }
  }

  /* Объект - линия */
  if (lLine) {
    for (pln = lLine; pln != NULL; pln = pln->next) {
      fprintf (fp, "%s %ld %f %f %f %f \n",
	       "LINE", pln->pixval, pln->p1->pt.x, pln->p1->pt.y, pln->p2->pt.x, pln->p2->pt.y);
    }
  }

  return (0);
}

enum obj {
  POINTt,
  LINEt,
}; 

struct {
  int fPOINT;
  int fLINE;
  int ulLINE;
} narg = {

  .fPOINT = 2,
  .fLINE  = 4,
  .ulLINE = 1,
};

/* Разбор строки. */
static void fs_parse (const char * buff)
{
# define CH_STR_PARAM 1 /* Количество строковых аргументов. */
# define F_NUM_PARAM 4 /* Предельное количество числовых аргументов типа "float". */
  int i, k;
  int obj; /* Объект */
  char tmps [CH_STR_PARAM] [TXT_MAX_STR]; /* аргументы типа "char *" */
  float f_num [F_NUM_PARAM]; /* аргументы типа "float" */
  unsigned long ul_num; /* аргумент типа "unsigned long" */
  int f_nparse = 0; /* Количество чисел типа "float" для распознания. */
  int ul_nparse = 0; /* Количество чисел типа "unsigned long" для распознания. */
  
  if (!(isalpha (buff [0])))
    return;

  /* Извлечение наименования команды. */
  /* sscanf (buff, "%s", &(tmps [0]));   */
  sscanf (buff, "%s", (tmps [0])); /* Changed */
  printf ("%s\n", (tmps [0]));

  /* Получение количества аргументов. */
  if ((strcmp (tmps [0], "POINT")) == 0) {
    f_nparse = narg.fPOINT ;
    obj = POINTt;
  }
  else if ((strcmp (tmps [0], "LINE")) == 0) {
    ul_nparse = narg.ulLINE;
    f_nparse = narg.fLINE;
    obj = LINEt;
  }
  else
    return;

  i = 0;
  if (ul_nparse == 1) { /* Первый аргумент типа "unsigned long" */
    i = fs_next_pos (i, buff);
    if (i < 0) goto resume;
    sscanf (&buff [i], "%ld", &ul_num);
    printf ("%ld ", ul_num);
    i = fs_scroll_str (i, buff);
    if (i < 0) goto resume;
  }  
  for (k=0; k<f_nparse; k++) { /* Аргументы типа "float" */
    i = fs_next_pos (i, buff);
    if (i < 0) goto resume;
    sscanf (&buff [i], "%f", &f_num [k]);
    printf ("%f ", f_num [k]);
    i = fs_scroll_str (i, buff);
    if (i < 0) goto resume;
  }

 resume:
  printf ("%d\n", f_nparse);

  switch (obj) {
  case POINTt:
    ref_point_add (f_num [0], f_num [1]);
    break;
  case LINEt:
    line_draw (&lLine, ul_num, f_num [0], f_num [1], f_num [2], f_num [3]);
    break;
  }
}

/* Перейти к следующему числу */
static int fs_scroll_str (int i, const char * buff)
{
  int c;

  c = buff [i];
  while ((c != '\0') && ((isdigit (c)) || c == '-' || c == '+' || c == '.')) {
    c = buff [i];
    i++;
  }
  i--;

  if (buff [i] == '\0') return (-1);

  return (i);
}

/* Поиск следующей цифры, точки, символов '+' или '-' */
static int fs_next_pos (int i, const char * buff)
{
  int c;

  c = buff [i];
  while ((c != '\0') && !(isdigit (c)) && c != '-' && c != '+' && c != '.') {
    c = buff [i];
    i++;
  }
  i--;

  if (buff [i] == '\0') return (-1);

  return (i);
}
