/* main.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#define  __USE_GNU
# define __USE_ISOC99
#include <stdlib.h>
#include <float.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_w_main.h"
#include "h_w_tool.h"
#include "h_w_command.h"
#include "h_w_common.h"
#include "h_main.h"
#include "h_com_console.h"
#include "h_coords.h"
#include "h_fs.h"
#include "h_o_cursor.h"


/* Names of video modes */
enum {
  M_640_480,
  M_800_600,
  M_1024_786,
  M_1280_1024,
};

static struct {
  int width :32;
  int height :32;
} VMode [] = { {640, 480}, {800, 600}, {1024, 768}, {1280, 1024} };

/* Параметры файла инициализации "config.ini" */
static TABLE_FILE_CONFIG tab_config [] = {
  {0, conf_int, "null"},

  {450, conf_uint, "m_x_pos"},
  {451, conf_uint, "m_y_pos"},
  {452, conf_int, "m_cursor_width"},
  {453, conf_int, "m_cursor_height"},    
  {470, conf_float, "m_scale_coef"},
  
  {500, conf_ptr_char, "t_font_name"},
  {520, conf_uint, "t_win_width"},
  {521, conf_uint, "t_win_height"},
  {522, conf_uint, "t_x_pos"},
  {523, conf_uint, "t_y_pos"},

  {540, conf_ptr_char, "c_font_name"},
  {550, conf_uint, "c_win_width"},
  {551, conf_uint, "c_win_height"},
  {552, conf_uint, "c_x_pos"},
  {553, conf_uint, "c_y_pos"},
  
  {-1, conf_int, "null"},
};

static int pr_argc;
static char ** pr_argv;

static void global_init (void);
static void parse_args (int argc, char * argv[]);
static int parse_config (void);
static int next_param (char * param);
static void set_param (char * param, char * value);


int main (int argc, char * argv [])
{
  /* Передать параметры в основную программу */
  pr_argc = argc;
  pr_argv = argv;

  LCPInit ();
  LCPGo (); /* Основная программа */

  return (0);
}

 
/* Инициализация данных.
   Отображение окон. */
int LCPInit (void)
{
  global_init (); /* Установить значения данных по умолчанию */
  parse_config (); /* Установить значения данных из файла "config.ini" */
  parse_args (pr_argc, pr_argv); /* Прочитать параметры командной строки */

  
  return (0);
}

/* Основная программа */
int LCPGo (void)
{
  if (XWndComCon) {
    pthread_create (&thread_id, NULL, &win_main, NULL); /* Отобразить главное окно */
    com_con (); /* Отобразить консоль */
  } else if (XWndWork) {
    win_main (NULL); /* Отобразить графическую консоль (по умолчанию). */
  } else if (XComCon) {
    fptr ? fs_read_file (fptr) : 0; /* Прочитать данные из файла */
    com_con (); /* Только системная консоль */
  } else {
    return (-1);
  }

  return (0);
}

/* Инициализация глобальных данных */
static void global_init (void)
{
  XWndWork = 1; /* "1" - по умолчанию */
  XWndComCon = 0; /* "0" */
  XComCon = 0; /* "0" */

/**
 * @variable: FILE * fptr
 * @files: "h_w_common.h", "fs.c"
 */
  fptr = NULL;

/**
 * @variable: NUMBERS nums
 * @files: "h_com_console.h", "fs.c"
 */
  nums.cnt = 0;
  memset (nums.ord, 0, sizeof (nums.ord));

  /* "h_coords.h" */
  coords_pan = coords_pan_no_lines;
  coords_pan_after = coords_pan_no_lines_after;
  m_scale_coef = 1.065;
  
  /* Установка шрифта */
  select_font = 0;
  
  /* Главное окно */
  m_x_pos = 0;
  m_y_pos = 0;
  CurMode = M_800_600;
  m_win_width = VMode [CurMode].width;
  m_win_height = VMode [CurMode].height;
  m_cursor_width = m_win_width;
  m_cursor_height = m_win_height;
  
  /* Окно "w_tool" */
  t_font_name = "-*-fixed-*-*-*-*-*-*-*-*-*-*-*-*";
  t_win_width = t_WIN_WIDTH;
  t_win_height = t_WIN_HEIGHT;
  t_x_pos = 0;
  t_y_pos = 0;

  /* Консоль "w_command" */
  c_font_name = "-*-fixed-*-*-*-*-*-*-*-*-*-*-*-*";
  c_nstr = 50;
  c_win_width  = c_WIN_WIDTH;
  c_win_height = c_WIN_HEIGHT;
  c_x_pos = 0;
  c_y_pos = 0;
}

/* Чтение параметров "config.ini" */
static int parse_config (void)
{
  FILE *fp;
  char param [MAXPARAM];
  char tmp_str [BUFSIZ];
  char *home_path;
  char *config_path = "/.config/LineColorPaint/config.ini";
  int str_hp_len, str_cp_len;

  home_path = getenv ("HOME");
  str_hp_len = strlen (home_path);
  str_cp_len = strlen (config_path);

  if ((str_hp_len + str_cp_len + 1) > BUFSIZ)
    return (-1);
    
  strcpy (tmp_str, home_path);
  strcat (tmp_str, config_path);

  if ((fp = fopen (tmp_str, "r")) == NULL) {
    if ((fp = fopen ("/usr/local/share/LineColorPaint/config.ini", "r")) == NULL) {
      if ((fp = fopen ("./config.ini", "r")) == NULL) {
	com_con_err_print ("config.ini is missing");    
	return (-1);
      }
    }
  }

  while (fgets (param, (MAXPARAM-1), fp) != NULL) {
    next_param (param);
  }

  fclose (fp);
  return (0);
}

/* Следующий параметр "config.ini" */
static int next_param (char * param)
{
  param [strlen (param) - 1] = '\0'; /* Удалить символ '\n' */

  char pbuf [MAXPARAM]; /* Содержит название параметра */
  char prm [MAXPARAM]; /* Содержит значение параметра */

  int i = 0;
  char ch = param [i];
  while (ch != '\0' && (ch == ' ' || ch == '\t')) { /* Пропустить лишние символы */
    ch = param [++i];
  }

  if (param [i] == ';') return (0); /* Это комментарий */
  if (!isalpha (param [i])) return (-1); /* В начале строки отсутствует название параметра */

  int j = 0;
  while (ch != '\0' && (ch != ' ' && ch != '\t')) { /* Записать название параметра */
    pbuf [j++] = ch = param [i++];
  }
  pbuf [j-1] = '\0';

  strcpy (prm, &param[i]);

  set_param (pbuf, prm); /* Установить параметр */

  return (0);
}

/* Установить параметр из "config.ini" */
static void set_param (char * param, char * value)
{
  value = com_con_no_blank (value); /* Удалить пробелы из строки */

  int i = 0;
  while (tab_config [i].id >= 0) {
    if ((strcmp (tab_config [i].name, param)) == 0)
      break;
    i++;
  }

  switch (tab_config [i].id) {
  case 450:
    m_x_pos = strtoul (value, NULL, 10);
    break;
  case 451:
    m_y_pos = strtoul (value, NULL, 10);
    break;
  case 452:
    m_cursor_width = strtol (value, NULL, 10);    
    break;
  case 453:
    m_cursor_height = strtol (value, NULL, 10);    
    break;    
  case 470:
    m_scale_coef = strtof (value, NULL);
    printf ("m_scale_coef: %f\n", m_scale_coef);    
    break;
  case 500:
    t_font_name = malloc ((strlen (value) + 1));
    strcpy (t_font_name, value);
    break;
  case 520:
    t_win_width = strtoul (value, NULL, 10);
    break;
  case 521:
    t_win_height = strtoul (value, NULL, 10);
    break;
  case 522:
    t_x_pos = strtoul (value, NULL, 10);
    break;
  case 523:
    t_y_pos = strtoul (value, NULL, 10);
    break;
  case 540:
    c_font_name = malloc ((strlen (value) + 1));
    strcpy (c_font_name, value);
    break;
  case 550:
    c_win_width = strtoul (value, NULL, 10);
    break;
  case 551:
    c_win_height = strtoul (value, NULL, 10);
    break;
  case 552:
    c_x_pos = strtoul (value, NULL, 10);
    break;
  case 553:
    c_y_pos = strtoul (value, NULL, 10);
    break;
  default:
    com_con_err_print ("set_param: BAD PARAMETER");
    break;
  }
}

/* Определение параметров командной строки.
   Имена параметров содержатся в структуре
   "tab_param". */
static void parse_args (int argc, char * argv[])
{
  int slen = 0;
  int n;
  int STRNG;
  
  STRNG=BUFF_SIZE-1;

  while (--argc > 0) {
    ++argv;
    slen = strlen (*argv);
    if (slen > STRNG) continue;

    strcpy (arg_buff, *argv);
    n = com_con_search_param (1, arg_buff);
    if (tab_param [n].num == (-1)) {
      if (!fptr) {
	if ((fptr = fs_open_file (arg_buff, "rw")) != NULL) {
	  file_name = strdup (arg_buff);
	  //printf ("%p %s\n", fptr, file_name);
	}
      }
      continue;
    }

    switch (tab_param [n].num) {
    case COM_NO_WINDOW:
      XWndWork = 0;
      XWndComCon = 0;
      XComCon = 1;
      break;
    case COM_WND_CON:
      XWndWork = 0;
      XComCon = 0;
      XWndComCon = 1;
      break;
    case COM_SELECT_FONT:
      select_font = 1;
      break;
    case COM_VERSION:
      com_con_print (tab_param [n].hlpmsg);
      exit (0);
      break;
    case COM_PAN_DRAW:
      coords_pan = coords_pan_draw_lines;
      coords_pan_after = coords_pan_draw_lines_after;
      break;
    default:
      break;
    }
  }
}
