/* o_cursor.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

/* Объект: курсор.
   Файл содержит функции для управления курсором.
   Объект относится к главному окну "w_main.c".
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>

#include "h_w_main.h"
#include "h_w_common.h"
#include "h_o_cursor.h"
#include "h_pix_cursor.h"


/* Установка данных курсора */
void cursor_init (void)
{
  CrCurs.flag = 0;
  CrCurs.sizeFlag = 0;
  CrCurs.X = 0;
  CrCurs.Y = 0;
  CrCurs.oldX = 0;
  CrCurs.oldY = 0;
  CrCurs.width = m_cursor_width;
  CrCurs.height = m_cursor_height;
  CrCurs.cursor = 29;

  if (CrCurs.width >= m_win_width)
    CrCurs.sizeFlag = 1;

  if (CrCurs.height >= m_win_height)
    CrCurs.sizeFlag = 1;

  if (CrCurs.width < 0 || CrCurs.width > m_win_width) {
    CrCurs.width = m_win_width;
    fprintf (stderr, "%s\n", "Cursor width: error size");
  }

  if (CrCurs.height < 0 || CrCurs.height > m_win_height){
    CrCurs.height = m_win_height;
    fprintf (stderr, "%s\n", "Cursor height: error size");
  }

  win_common_set_color (&CrCurs.clr, 65535, 65535, 65535);

  CrCurs.pixmap = XCreateBitmapFromData (display, m_window, (char *) cursor_bits,
					 cursor_width, cursor_height);
  CrCurs.cursor = XCreatePixmapCursor (display, CrCurs.pixmap, CrCurs.pixmap,
				       &CrCurs.clr, &CrCurs.clr, 1, 1);
  if (CrCurs.cursor == 0) {
    CrCurs.cursor = XCreateFontCursor (display, 29);
    XDefineCursor (display, m_window, CrCurs.cursor);
  } else {
    XDefineCursor (display, m_window, CrCurs.cursor);
  }

  if (CrCurs.sizeFlag)
    cursor_draw = cursor_draw_size_flag_1;
  else
    cursor_draw = cursor_draw_size_flag_0;
}

/* Рисование курсора (Размеры курсора не установлены)
   Курсор занимает весь экран */
void __attribute__ ((__always_inline__))
cursor_draw_size_flag_1 (void)
{
  XSetForeground (display, m_gc, CrCurs.clr.pixel);
  win_main_change_mask (GXxor);

  if (CrCurs.flag) {
    XDrawLine (display, m_pxm, m_gc, CrCurs.oldX, 0, CrCurs.oldX, m_win_height);
    XDrawLine (display, m_pxm, m_gc, 0, CrCurs.oldY, m_win_width, CrCurs.oldY);
  }
  /* Сохранение позиции курсора в "CrCurs.oldX", "CrCurs.oldY"  и
     установка флага изменения. */
  XDrawLine (display, m_pxm, m_gc, CrCurs.oldX = CrCurs.X,
	     0,
	     CrCurs.X,
	     m_win_height);
  XDrawLine (display, m_pxm, m_gc, 0,
	     CrCurs.oldY = CrCurs.Y,
	     m_win_width,
	     CrCurs.Y);

  CrCurs.flag = 1;
}

/* Рисование курсора (Размеры курсора установлены)
   Курсор часть экрана */
void __attribute__ ((__always_inline__))
cursor_draw_size_flag_0 (void)
{
  XSetForeground (display, m_gc, CrCurs.clr.pixel);
  win_main_change_mask (GXxor);

  if (CrCurs.flag) {
    XDrawLine (display, m_pxm, m_gc, CrCurs.oldX, CrCurs.oYtop, CrCurs.oldX, CrCurs.oYbot);
    XDrawLine (display, m_pxm, m_gc, CrCurs.oXlef, CrCurs.oldY, CrCurs.oXrig, CrCurs.oldY);
  }
  /*	Сохранение позиции курсора в "CrCurs.oldX", "CrCurs.oldY"  и
	установка флага изменения. */
  XDrawLine (display, m_pxm, m_gc, CrCurs.oldX = CrCurs.X,
	     CrCurs.oYtop = CrCurs.Y+CrCurs.height/2,
	     CrCurs.X,
	     CrCurs.oYbot = CrCurs.Y-CrCurs.height/2);
  XDrawLine (display, m_pxm, m_gc, CrCurs.oXlef = CrCurs.X-CrCurs.width/2,
	     CrCurs.oldY = CrCurs.Y,
	     CrCurs.oXrig = CrCurs.X+CrCurs.width/2,
	     CrCurs.Y);

  CrCurs.flag = 1;
}

/* Скрыть курсор */
void cursor_hide (void)
{
  float tmpX, tmpY;
  tmpX = CrCurs.X;
  tmpY = CrCurs.Y;

  CrCurs.X = -10;
  CrCurs.Y = -10;
  cursor_draw ();

  CrCurs.X = tmpX;
  CrCurs.Y = tmpY;

# if 0
  win_main_change_mask (GXcopy);
  XCopyArea (display, m_pxm, m_window, m_gc, 0, 0, m_win_width, m_win_height, 0, 0);
# endif
}
