/* point.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

/* Создание объекта - точка.
   Объект точка применяется для создания объектов:
   линия, полилиния, прямоугольник, дуга, окружность, кривая.
   Непосредственно на экране не отображается.
   
   Основные параметры объекта:
   1) дескриптор (dsc),
   2) координаты (pt - мировые, pp - оконные),
   3) предыдущий узел (* prev),
   4) следующий узел (* next).

   Поиск осуществляется:
   1) по дескриптору,
   2) по координатам.   */

#include <stdio.h>
#include <stdlib.h>

#include "h_point.h"
#include "h_coords.h"

#define TRACE_PROG_POINT 0


/* Добавление точки в список в порядке увеличения
   величины дескриптора.
   Функция возвращает указатель на вставленный
   элемент списка. */
pPL point_add (register pPL * p, float x, float y)
{
  static unsigned long dsc = 0; /* Дескриптор. */
  register pPL new; /* Новый элемент списка. */

  if ((new = malloc (sizeof (PL))) == NULL) return (0);
  new->dsc = dsc++;
  new->pt.x = x; new->pt.y = y;
  new->pp.x = x * m_scale + zeroPt.x; new->pp.y = y * m_scale + zeroPt.y; /* Пересчет из мировых координат в экранные. */
#if TRACE_PROG_POINT
  printf ("%f %f \n", zeroPt.x, zeroPt.y); fflush(stdout);
#endif
  /* Если "p" == NULL (список пуст). */
  if (*p == NULL) {
    new->prev = NULL;
    new->next = NULL;
    *p = new;
    return (new);
  }

  new->prev = NULL;
  new->next = *p;
  (*p) = new;

# if 0
  register pPL pt;
  for (pt = *p; pt != NULL; pt = pt->next)
    printf ("%ld  %f  %f \n", pt->dsc, pt->pt.x, pt->pt.y);
  fflush (stdout);
# endif
  return (new);
}

/* Удаление точки с соответствующим дескриптором
   из списка.
   Функция возвращает указатель на начало списка. */
pPL point_delete (pPL * p, unsigned long dsc)
{
  if (*p == NULL) return (0);
  pPL head = *p;

  for ( ; *p != NULL; *p = (*p)->next)
    if ((*p)->dsc == dsc) {
      if ((*p)->prev)
      	(*p)->prev->next = (*p)->next;
      else if ((*p)->next)
	head = (*p)->next;

      if ((*p)->next) (*p)->next->prev = (*p)->prev;

      free (*p);
      if ((*p)->next==NULL && (*p)->prev==NULL) {
	head = *p = NULL;
	return (head);
      }
    }

  return (head);
}

/* Просмотр всех элементов списка.
   pPL - указатель на начало списка. */
void point_print_test (pPL * ptr)
{
  register pPL p;

  for (p = *ptr; p != NULL; p = p->next)
    printf ("%ld  %f  %f \n", p->dsc, p->pt.x, p->pt.y);
  fflush (stdout);
}
