/* reference_point.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

/*
   Создание точек для ориентирования на плоскости.
*/


#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>

#include "h_LCP_reference_point.h"
#include "h_coords.h"
#include "h_o_cursor.h"

static void ref_point_del (pRefPointList ptr);
static pRefPointList ref_point_lookup (float x, float y);
static pRefPointList ref_point_lookup2 (float x, float y);



/* Добавление точки в отсортированный список.
   Если точка с соответствующими координатами
   уже есть, то новая в любом случае добавляется. */
void ref_point_add (float x, float y)
{
  register pRefPointList new;
  register pRefPointList ptr;
  int flag = 0;

  if ((new = malloc (sizeof (RefPointList))) == NULL) return;
  new->pt.x = x; new->pt.y = y;
  /* Пересчет из мировых координат в экранные */  
  new->pp.x = x * m_scale + zeroPt.x; new->pp.y = y * m_scale + zeroPt.y;

  /* Если "refPoint" == NULL (список пуст). */
  if (refPoint == NULL) {
    new->prev = NULL;
    new->next = NULL;
    refPoint = new;
    return;
  }

  /* Поиск места для нового узла,
     вставка узла в отсортированный
     список. */
  ptr = refPoint;
  while (1) {
    if (ptr->pt.x > x) {
      new->prev = ptr->prev;
      new->next = ptr;
      if (ptr->prev) ptr->prev->next = new;
      ptr->prev = new;
      if (ptr == refPoint) refPoint = new;
      return;
    }

    if ((ptr->pt.x == x) || flag) {
      if (ptr->pt.y < y) {
	flag = 1;
      }
      else {
	new->prev = ptr->prev;
	new->next = ptr;
	if (ptr->prev) ptr->prev->next= new;
	ptr->prev = new;
	if (ptr == refPoint) refPoint = new;
	return;
      }
    }

    if (ptr->next == NULL) {
      new->prev = ptr;
      new->next = NULL;
      ptr->next = new;
      return;
    }

    ptr = ptr->next;
  }

}

/* Удаление элемента из списка */
static void ref_point_del (pRefPointList ptr)
{
  if (ptr == NULL) return;
  if (ptr->prev)
    ptr->prev->next = ptr->next;
  else
    if (ptr->next) refPoint = ptr->next;

  if (ptr->next) ptr->next->prev = ptr->prev;

  free (ptr);

  if (ptr->next==NULL && ptr->prev==NULL) refPoint=NULL;
}

/* Удаление точки с координатами ("X", "Y") */
void ref_point_dl (float x, float y)
{
  pRefPointList ptr;
  ptr = ref_point_lookup (x, y);
  ref_point_del (ptr);
}

/* Удаление всех элементов списка */
void ref_point_del_all (void)
{
  register pRefPointList ptr;
  register pRefPointList next;

  ptr = refPoint;
  while (ptr != NULL) {
    next = ptr->next;
    free (ptr);
    ptr = NULL;
    ptr = next;
  }
  refPoint = NULL;
}

/* Просмотр всех элементов списка */
void ref_point_printList (void)
{
  register pRefPointList ptr;
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next)
    printf ("%f  %f \n", ptr->pt.x, ptr->pt.y);
  fflush (stdout);
}

/* Проверка элементов "next" и "prev" в списке */
void ref_point_print2List (void)
{
  register pRefPointList ptr;
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next)
    printf ("%f  %f \n", ptr->pt.x, ptr->pt.y);

  for (ptr = refPoint; ptr != NULL; ) {
    if (ptr->next == NULL) break;
    ptr = ptr->next;
  }

  for ( ; ptr != NULL; ptr = ptr->prev)
    printf ("%f  %f \n", ptr->pt.x, ptr->pt.y);

  fflush (stdout);
}

/* Поиск элемента в списке с координатами
   "x" и "y". */
static pRefPointList ref_point_lookup (float x, float y)
{
  register pRefPointList ptr;
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next)
    if ((ptr->pt.x == x) && (ptr->pt.y == y)) {
      return (ptr);
    }

  return (NULL);
}

/* Поиск элемента в списке с координатами "x" и "y".
   Функция использует размер указателя курсора "POINTER_HALF_SIZE".
   Функция возвращает указатель на найденную точку, или "NULL",
   если точка не найдена.
*/
static pRefPointList ref_point_lookup2 (float x, float y)
{
  register pRefPointList ptr;
  int xmin = x - POINTER_HALF_SIZE;
  int xmax = x + POINTER_HALF_SIZE;
  int ymin = y - POINTER_HALF_SIZE;
  int ymax = y + POINTER_HALF_SIZE;
  
  for (ptr = refPoint; ptr != NULL; ptr = ptr->next) {
    if (!((ptr->pt.x<xmin) << 3 | (ptr->pt.x>xmax) << 2 |
	  (ptr->pt.y<ymin) << 1 | (ptr->pt.y>ymax)))
      return (ptr);    
  }

  return (NULL);
}


#define ref_point_search_cycle() ({			  \
   {							  \
    if (!((ptr->pt.x<xmin) << 3 | (ptr->pt.x>xmax) << 2 | \
	  (ptr->pt.y<ymin) << 1 | (ptr->pt.y>ymax))) {	  \
      cur_pt = ptr->next;				  \
      return (ptr);					  \
    }							  \
   }})
/* Поиск элемента в списке с координатами "x" и "y".
   Функция использует размер указателя курсора "POINTER_HALF_SIZE".
   Функция возвращает указатель на найденную точку, или "NULL",
   если точка не найдена. */
pRefPointList ref_point_lookup_next (float x, float y)
{
  static pRefPointList cur_pt = NULL;

  if (refPoint == NULL) return (NULL);
  if (cur_pt == NULL) cur_pt = refPoint;
  
  register pRefPointList ptr;
  int xmin = x - POINTER_HALF_SIZE;
  int xmax = x + POINTER_HALF_SIZE;
  int ymin = y - POINTER_HALF_SIZE;
  int ymax = y + POINTER_HALF_SIZE;

  /* Поиск от текущей точки до конца списка */
  for (ptr = cur_pt; ptr != NULL; ptr = ptr->next)
    ref_point_search_cycle();

  /* Поиск от начала списка до текущей точки */  
  for (ptr = refPoint; ptr != cur_pt && ptr != NULL; ptr = ptr->next)
    ref_point_search_cycle();
  
  return (NULL);
}
