/* w_command.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_LCP_global.h"
#include "h_w_main.h"
#include "h_w_command.h"
#include "h_w_common.h"
#include "h_o_cursor.h"
#include "h_coords.h"
#include "h_com_console.h"
#include "h_ico_color.h"


#define win_com_change_mask(f1) win_common_change_mask(c_gc, c_valuemask, c_values, f1)
#define RCoorX(x) (x)
#define RCoorY(y) (display_height - y)
#define VBLANK 1 /* Межстрочный интервал */
#define HBLANK 2
#define c_WIN_BORDER 2

static char prm [] = ">";
static char hq	 [] = "?";
static char mod [] = "!";
static char mode_word [] = "Mode: ";
static char scale_word [] = " || Scale: ";
static RECT t_rect = {10, 0, c_WIN_WIDTH, 50};
static RECT h_rect = {10, 0, c_WIN_WIDTH, 50};
static RECT m_rect = {10, 0, c_WIN_WIDTH, 50};
static RECT * b_rect;
static char * c_wName = "Console";
static unsigned long c_valuemask = 0;
static int cpos = 0; /* Позиция курсора */
static long int c_event_mask;
static XFontStruct * comFI;
static XFontStruct * c_foinf;
static XWindowAttributes cAttr;
static XColor con_bg_clr; /* Цвет консоли */
static XColor con_font_clr; /* Цвет шрифта */
static unsigned int con_width, con_height;
static XGCValues c_values;
static Pixmap c_pxm;
static GC c_gc;
static int iCmd; /* Индекс команды в массиве */
/* История комманд */
#define HIST_LINES 15 /* Количество строк в истории */
static int currHistLine;
static char (* hist_bfs) [TXT_MAX_STR]; /* Массив строк */
/* Буфферные строки */
static char (* con_bfs) [TXT_MAX_STR]; /* Массив строк */
static LABEL * cBfs; /* Указатель на массив буфферных строк (без курсора). */
static char * tmp_s_ptr; /* Временный указатель для обмена элементов "cBfs" строками (функция "con_ret"). */
/* Строка режима */
static char con_mod [TXT_MAX_STR];
static LABEL cMod;
/* Строка с подсказкой */
static char con_hlp [TXT_MAX_STR];
static LABEL cHlp;
/* Командная строка */
static TEXT cTxt;

static void win_com_con_ret (void); /* Выполнить команду, прокрутить строки вверх на одну позицию */
static void win_com_parse_command (void); /* Разбор параметров строки */
static void win_com_console_set (void); /* Устанавливает параметры консоли перед первым запуском */
static void win_com_hlp_def (void); /* Строка c подсказкой по умолчанию */
static void win_com_hist_next (void);
static void win_com_hist_prev (void);


int win_com (void)
{
  Atom pnprotocol[2];
  Atom wmprotocols;
  int depth;

  c_event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask |
    ButtonReleaseMask | EnterWindowMask | LeaveWindowMask | PointerMotionMask
    | Button1MotionMask | Button2MotionMask | Button3MotionMask |
    Button4MotionMask | Button5MotionMask | ButtonMotionMask | KeymapStateMask
    | ExposureMask | VisibilityChangeMask | SubstructureNotifyMask |
    SubstructureRedirectMask | FocusChangeMask | PropertyChangeMask |
    ColormapChangeMask | OwnerGrabButtonMask;
 
  /* Создать окно, установить цвет и прорисовать */
  c_window = XCreateWindow (display, RootWindow (display, screen),
			    0, 0, c_win_width, c_win_height, c_WIN_BORDER,
			    CopyFromParent, CopyFromParent,
			    CopyFromParent, 0, 0);
  XMapWindow (display, c_window);
  XMoveWindow (display, c_window, c_x_pos, c_y_pos);
  XFlush (display);
  XSelectInput (display, c_window, c_event_mask);
  XEventsQueued (display, QueuedAlready);

  /* Получение графического контекста */
  c_valuemask = GCFunction | GCLineWidth;
  c_values.line_width = 3;
  c_values.function = GXcopy;
  depth = DefaultDepth (display, screen);
  c_pxm = XCreatePixmap (display, c_window, c_win_width, display_height, depth);
  if ((c_gc = XCreateGC (display, c_pxm, c_valuemask, &c_values)) < 0) {
    fprintf(stderr, "cannot create GC \n"); exit(1);
  }
  win_common_set_color (&con_bg_clr, 53000, 53000, 53000);
  XSetWindowBackground (display, c_window, con_bg_clr.pixel);

  /* Установка фона для "c_pxm" */
  win_common_change_mask (c_gc, c_valuemask, c_values, GXcopy);
  XSetForeground (display, c_gc, con_bg_clr.pixel);
  XFillRectangle (display, c_window, c_gc, 0, 0, c_win_width, display_height);
  XFillRectangle (display, c_pxm, c_gc, 0, 0, c_win_width, display_height);

  /* Установки менеджера окон */
  pnprotocol[0] = XInternAtom (display, "WM_TAKE_FOCUS", True);
  pnprotocol[1] = XInternAtom (display, "WM_DELETE_WINDOW",True);
  wmprotocols = XInternAtom(display, "WM_PROTOCOLS", True);
  XSetWMProtocols (display, c_window, pnprotocol, 2);

  /* Установить заголовок окна */
  XStoreName (display, c_window, c_wName);
  win_common_set_color_window_icon (c_window, LCP_color_xpm, icon_width, icon_height);

  /* Установить шрифт окна "c_font_name" */
  if (select_font) {
    if ((c_foinf = XLoadQueryFont (display, c_font_name))) { XSetFont (display, c_gc, c_foinf->fid); }
  }
  
  /* Установить параметры консоли */
  win_com_console_set ();

  /* Обновить окно */
  win_com_draw ();

  return(0);
}


#define MIN_CNT_BFS   5
#define MAX_CNT_BFS   50
/* Установить параметры консоли */
static void win_com_console_set (void)
{
  win_common_change_mask (c_gc, c_valuemask, c_values, GXcopy);
  win_common_set_color (&con_font_clr, 0, 0, 0);
  XSetForeground (display, c_gc, con_font_clr.pixel);

  con_width = c_win_width;
  int h = t_rect.height;
  int npos;
  nCmd = 0;   /* Номер текущей команды */
  iCmd = 0;   /* Индекс команды в массиве */

  if (c_nstr < MIN_CNT_BFS) c_nstr = MIN_CNT_BFS;
  if (c_nstr > MAX_CNT_BFS) c_nstr = MAX_CNT_BFS;

  GContext gcnt = XGContextFromGC (c_gc);
  comFI =  XQueryFont (display, gcnt);
  //h = comFI->ascent + comFI->descent + 2*MARGIN;
  h = label_get_font_height (c_gc);

# define VDESC 4
  /* Командная строка */
  int sl = strlen (prm);
  int twidth = XTextWidth (comFI, prm, sl);
  npos = h + VBLANK;
  t_rect.x = twidth + 2*HBLANK;
  t_rect.y = RCoorY(0) - npos;
  XDrawString (display, c_pxm, c_gc, HBLANK, RCoorY(0)-VBLANK - VDESC, prm, sl);
  text_create (display, c_pxm, screen, c_gc, con_str, t_rect, &cTxt);
  cTxt.cpos = cpos;
  text_write (display, c_pxm, c_gc, &cTxt);
  /* Строка с подсказкой */
  sl = strlen (hq);
  twidth = XTextWidth (comFI, hq, sl);
  h_rect.x = twidth + 2*HBLANK;
  h_rect.y = RCoorY(0) - 2*npos;
  XDrawString (display, c_pxm, c_gc, HBLANK, RCoorY(0) - VDESC -2*VBLANK-h, hq, sl);
  label_create (display, c_pxm, screen, c_gc, con_hlp, &h_rect, &cHlp);
  win_com_hlp_def ();
  label_draw (display, c_pxm, c_gc, &cHlp);
  /* Строка режима */
  sl = strlen (mod);
  twidth = XTextWidth (comFI, mod, sl);
  m_rect.x = twidth + 2*HBLANK;
  m_rect.y = RCoorY(0) - 3*npos;
  XDrawString (display, c_pxm, c_gc, HBLANK, RCoorY(0) - VDESC - 4*VBLANK-2*h, mod, sl);
  label_create (display, c_pxm, screen, c_gc, con_mod, &m_rect, &cMod);
  win_com_mode_def ();
  label_draw (display, c_pxm, c_gc, &cMod);
  /* История комманд */
  currHistLine = 0;
  hist_bfs = malloc (HIST_LINES*TXT_MAX_STR);
  memset (hist_bfs, 0, (HIST_LINES*TXT_MAX_STR));
  int i;
  for (i = 0; i < HIST_LINES; i++) {
    hist_bfs [i] [0] = '\0';
  }
  /* Буфферные строки */
  con_bfs  = malloc (c_nstr*TXT_MAX_STR);
  memset (con_bfs, 0, (c_nstr*TXT_MAX_STR));
  b_rect = malloc (c_nstr*sizeof (RECT));
  cBfs = (LABEL *) malloc (c_nstr*(sizeof (LABEL)));
  memset (cBfs, 0, (c_nstr*(sizeof (LABEL))));

  for (i = 0; i < c_nstr; i++) {
    b_rect [i].x = 0; b_rect [i].y = 0; b_rect [i].width = c_WIN_WIDTH; b_rect [i].height = 50;
  }
  for (i = 0; i < c_nstr; i++) {
    con_bfs [i] [0] = '\0';
  }
  for (i = 0; i < c_nstr; i++) {
    b_rect [i].y = RCoorY(0) - ((i+4)*npos + 2*VBLANK);
    cBfs [i].flags = NO_FRAME;
    label_create (display, c_pxm, screen, c_gc, con_bfs [i], &b_rect [i], &cBfs [i]);
    label_draw (display, c_pxm, c_gc, &cBfs [i]);
  }

  con_height = (i+4)*npos+2*VBLANK;

  IsParam = 0;
}

/* Основная функция вывода графики */
void win_com_draw (void)
{
  XGetWindowAttributes (display, c_window, &cAttr);
  c_win_height = cAttr.height;

  win_com_change_mask (GXcopy);

  XCopyArea ( display, c_pxm, c_window, c_gc, 0, display_height-con_height,
	      con_width, con_height, 0, c_win_height-con_height);
}

/* Выполнить команду */
static void win_com_con_ret (void)
{
  int i;
  tmp_s_ptr = cBfs [c_nstr-1].text;
  for (i = (c_nstr-1); i > 0; i--) {
    cBfs [i].text = cBfs [i-1].text;
    label_draw (display, c_pxm, c_gc, &cBfs [i]);
  }

  cBfs [0].text = tmp_s_ptr;
  strcpy (cBfs [0].text, con_str);

  for (i = (HIST_LINES-1); i > 0; i--) {
    strcpy (hist_bfs [i], hist_bfs [i-1]);
  }
  strcpy (hist_bfs [1], con_str);
  currHistLine = 0;

  text_write (display, c_pxm, c_gc, &cTxt);
  label_draw (display, c_pxm, c_gc, &cHlp);
  label_draw (display, c_pxm, c_gc, &cMod);
  label_draw (display, c_pxm, c_gc, &cBfs [0]);
}

/* Разбор параметров строки */
static void win_com_parse_command (void)
{
  if (IsParam) { /* IsParam = PARAM (Это параметры) */
    strcpy (num_buff, con_str);
    cursor_hide ();
    com_con_do_list (iCmd);
    win_com_hlp_def ();
    iCmd = 0;
    win_com_mode_def ();

    IsParam = COMMAND;
    return;
  }

  /* IsParam == COMMAND (Это команда) */
  char tS [TXT_MAX_STR]; /* Временная строка */
  strcpy (tS, con_str);

  com_con_no_blank (con_str);
  if (strlen (con_str) <= SHORT_PAR_LEN) {
    iCmd = com_con_search_param (0, con_str); /* поиск в списке коротких параметров */
  } else {
    iCmd = com_con_search_param (1, con_str); /* поиск в списке длинных параметров */
  }
#if 0
  printf ( "%d ", i);
  fflush (stdout);
#endif
  /* Команда не найдена, вернуть строку в буфер "con_str". */
  if ((tab_param [iCmd].num) == (-1)) {
    strcpy (con_str, tS);
    return;
  }

  nCmd = tab_param [iCmd].num;

  if ((strlen (tab_param [iCmd].hlpmsg)) < 80)
    strcpy (con_hlp, tab_param [iCmd].hlpmsg);
  else
    com_con_err_print ("win_com_parse_command: error");

  win_com_mode_def ();

  IsParam = PARAM;  /* Ввести параметры */
}

/* 	Функция получает несколько чисел,
 	проверяет корректность ввода, допустимые пределы.
 	В случае ввода неверных данных функция выводит
 	сообщение об ошибке.
*/
int win_com_get_nums (int cnt)
{
  int state = OUT;
  int i = 0;
  char ch;
  char lbuff [LNUM_MAX];
  int lnum = 0;
  int order = 0;

  if (cnt >= 1 && cnt < 5)
    nums.cnt = cnt;
  else {
    com_con_err_print ("error: parameters incorrect");
    return (-1);
  }

  while ((ch = num_buff [i]) != '\0' ) {
    if (isdigit (ch) || ch == '-' || ch == '.' || ch == ',' ) {
      if (ch == ',') ch = '.';
      if (state == OUT) {
	switch (order) {
	case 0:
	  lnum = 0;
	  order++;
	  break;
	case 1:
	case 2:
	case 3:
	case 4:
	  lbuff[lnum] = '\0';
	  nums.ord [order-1] = atof (lbuff);
	  lnum = 0;
	  order++;
	  break;
	default:
	  com_con_err_print ("error: too many parameters");
	  return (-1);
	}
      }
      lbuff [lnum] = ch;
      lnum++;
      if (lnum > LNUM_MAX-1) {
	com_con_err_print ("error: too long number");
	return (-1);
      }
      state = IN;
    } else {
      state = OUT;
    }

    i++;
  } // while

  lbuff [lnum] = '\0';
  nums.ord [order-1] = atof (lbuff);

  if (order < 1) {
    com_con_err_print ("error: too little parameters");
    return (-1);
  }

  return (0);
}

/* Установить элемент управления "cTxt" текущим  */
void win_com_set_focus (void)
{
  pTcr = &cTxt;
}

/* Обновить командную строку */
void win_com_refresh (void)
{
  text_write (display, c_pxm, c_gc, pTcr);
}

/* Изменить позицию курсора в консоли.
   Переопределение функции "text_change_pos" */
void win_com_change_pos (int pos)
{
  text_change_pos (pos);
}

/* Обработчик событий клавиатуры */
void win_com_key_release (XEvent * ev, char * buff_kb, KeySym keysym)
{
  win_com_set_focus ();
  text_key_release (ev, buff_kb, keysym);
  win_com_refresh ();
}

/* Следующая команда. */
static void win_com_hist_next (void) {
  currHistLine--;
  if (currHistLine < 0) currHistLine = 0;
  strcpy (con_str, hist_bfs [currHistLine]);
  win_com_refresh ();
  win_com_draw ();
}

/* Предыдущая команда. */
static void win_com_hist_prev (void) {
  currHistLine++;
  if (currHistLine > HIST_LINES-1) currHistLine = HIST_LINES-1;
  strcpy (con_str, hist_bfs [currHistLine]);
  win_com_refresh ();
  win_com_draw ();
}

/* Обработчик событий клавиатуры */
void win_com_key_press (XEvent * ev, char * buff_kb, KeySym keysym)
{
  win_com_set_focus ();

  switch (keysym){
  case XK_Return:
  case 0xff8d: /* Enter на дополнительной клавиатуре */
    win_com_parse_command ();
    win_com_con_ret ();
    break;
  case XK_Up:
    win_com_hist_prev ();
    break;
  case XK_Down:
    win_com_hist_next ();
    break;
  case 0x006a: /* XK_j */
    if (fl_control) {
      win_com_parse_command ();
      win_com_con_ret ();
    }
    break;
  case 0x006e: /* XK_n */
    if (fl_control) {
      win_com_hist_next ();
      return;
    }
    break;
  case 0x0070: /* XK_p */
    if (fl_control) {
      win_com_hist_prev ();
      return;
    }
    break;
  default:

    break;
  }

  text_key_press (ev, buff_kb, keysym);
  win_com_refresh ();
  win_com_draw ();
}

/* Обработчик событий мыши */
void win_com_button_press (int x, int y)
{
  win_com_set_focus ();
  text_button_press (x, y);
  win_com_refresh ();
  win_com_draw ();
}

/* Строка c подсказкой по умолчанию */
static void win_com_hlp_def (void)
{
  if ((strlen (tab_param [0].hlpmsg)) < 80)
    strcpy (con_hlp, tab_param [0].hlpmsg);
  else
    com_con_err_print ("win_com_parse_command: error");
}

/* Строка режима по умолчанию */
void win_com_mode_def (void)
{
  char ts [20];
  sprintf (ts, "%f", m_scale);

  if (((strlen (tab_param [iCmd].param [1]))+(strlen (mode_word))  + (strlen (scale_word)) +
       (strlen (ts))) < 80) {
    strcpy (con_mod, mode_word);
    strcat (con_mod, tab_param [iCmd].param [1]);
    strcat (con_mod, scale_word);
    strcat (con_mod, ts);
  } else
    com_con_err_print ("win_com_parse_command: error");

  label_draw (display, c_pxm, c_gc, &cMod);
  win_com_draw ();
}

/* Освободить память */
void win_com_mem_free (void)
{
  XFreeFont (display, comFI);
  XFreeFont (display, c_foinf);
  XFreeGC (display, c_gc);
  free (con_bfs);
  free (cBfs);
}
