/* w_common.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>

#include "h_w_common.h"
#include "h_w_main.h"


/* Перерисовать окно */
void win_common_redraw_window (Window window)
{
  XEvent xEv;
  xEv.type = Expose;
  XSendEvent (display, window, 0, ExposureMask, &xEv);
  XFlush(display);
}

/* Получить цвет из переменной "pixval" */
void win_common_get_RGB_val (unsigned long pixval, unsigned long * R, unsigned long * G, unsigned long * B)
{
  *B = (pixval &  0xff); /* Получить синий цвет */
  *G = ((pixval & 0xff00) >> 8); /* Получить зеленый цвет */
  *R = ((pixval & 0xff0000) >> 16); /* Получить красный цвет */
}

/* Установить цвет */
void win_common_set_color (XColor * color, unsigned short _r, unsigned short _g, unsigned short _b)
{
  Colormap cm;

  color->red = _r;
  color->green = _g;
  color->blue = _b;
  color->flags = DoRed | DoGreen | DoBlue;

  cm = DefaultColormap(display, 0);
  XAllocColor (display, cm, color);
}

/* Изменить маску */
void win_common_change_mask (GC gc, unsigned long valmask, XGCValues val, int func)
{
  valmask = GCFunction;
  val.function = func;
  XChangeGC (display, gc, valmask, &val);
}

/* Изменить толщину линии */
void win_common_change_line_width (GC gc, unsigned long valmask, XGCValues val, int width)
{
  valmask = GCLineWidth;
  val.line_width = width;
  XChangeGC (display, gc, valmask, &val);
}

/* Получить параметры шрифта */
int win_common_get_font_info (GC gc, int * asc, int * desc, int * dir)
{
  GContext gcnt = XGContextFromGC (gc);
  char * str = "string";
  XCharStruct xcs;
  XQueryTextExtents (display, gcnt, str, strlen(str), dir, asc, desc, &xcs);

  return (0);
}

/* Загрузка курсора */
int win_common_load_cursor (Window window, char * FileName, Pixmap * pixmap)
{
  int rc;
  unsigned int curW = 3;
  unsigned int curH = 3;

  rc = XReadBitmapFile (display, window, FileName,
			&curW, &curH, pixmap, 0, 0);
  switch (rc){
  case BitmapOpenFailed:
    fprintf (stderr, "win_common_load_cursor - dont have open file  \n");
    break;
  case BitmapFileInvalid:
    fprintf (stderr, "win_common_load_cursor - file is not correct \n");
    break;
  case BitmapNoMemory:
    fprintf (stderr, "win_common_load_cursor - dont have a memory \n");
    break;
  case BitmapSuccess:
    /*fprintf (stderr, "win_common_load_cursor - file load \n");*/
    return (0);
    break;
  default:
    break;
  }
  return (-1);
}

/* Загрузка цветного изображения в формате "xpm" */
int win_common_load_img (char * FileName, XImage ** xi)
{
  int rc;

  rc = XpmReadFileToImage (display, FileName, xi, 0, 0);
  switch (rc){
  case XpmOpenFailed:
    fprintf (stderr, "win_common_load_img - dont have open file \n");
    break;
  case XpmFileInvalid:
    fprintf (stderr, "win_common_load_img - file is not correct \n");
    break;
  case XpmNoMemory:
    fprintf (stderr, "win_common_load_img - dont have a memory \n");
    break;
  case XpmColorError:
    fprintf (stderr, "win_common_load_img - color error \n");
    break;
  case XpmColorFailed:
    fprintf (stderr, "win_common_load_img - color filed \n");
    break;
  case XpmSuccess:
    /* fprintf (stderr, "win_common_load_img - file load \n"); */
    return (0);
    break;
  default:
    break;
  }
  return (-1);
}

/* Загрузка цветного изображения в формате "xpm" из данных */
int win_common_load_img_from_data (char ** data, XImage ** xi)
{
  int rc;
  rc = XpmCreateImageFromData (display, data, xi, 0, 0);
  switch (rc){
  case XpmOpenFailed:
    fprintf (stderr, "win_common_load_imgFD - dont have open file \n");
    break;
  case XpmFileInvalid:
    fprintf (stderr, "win_common_load_imgFD - file is not correct \n");
    break;
  case XpmNoMemory:
    fprintf (stderr, "win_common_load_imgFD - dont have a memory \n");
    break;
  case XpmColorError:
    fprintf (stderr, "win_common_load_imgFD - color error \n");
    break;
  case XpmColorFailed:
    fprintf (stderr, "win_common_load_imgFD - color filed \n");
    break;
  case XpmSuccess:
    /* fprintf (stderr, "win_common_load_imgFD - file load \n"); */
    return (0);
    break;
  default:
    break;
  }
  return (-1);
}

/* Установить "pixmap" в заголовок окна */
void win_common_set_window_icon (Window win, unsigned char * icon, int width, int height)
{
  XWMHints* win_hints;
  Pixmap icon_pixmap;

  if (!(win_hints = XAllocWMHints())) { fprintf(stderr, "win_common_set_window_icon: error\n"); exit(1); }

  win_hints->flags = StateHint | IconPositionHint | IconPixmapHint;

  if (!(icon_pixmap = XCreateBitmapFromData(display, win, (char *) icon, width, height))) {
    fprintf(stderr, "win_common_set_window_icon: error\n"); exit(1);
  }

  win_hints->initial_state = IconicState;
  win_hints->icon_pixmap = icon_pixmap;
  win_hints->icon_x = 0; win_hints->icon_y = 0;

  XSetWMHints(display, win, win_hints);

  XFree(win_hints);
}

/* Установить "pixmap" в заголовок окна */
void win_common_set_color_window_icon (Window win, char ** icon, int width, int height)
{
  XWMHints* win_hints;
  Pixmap icon_pixmap;
  Pixmap mask_pixmap;
  int status;

  if (!(win_hints = XAllocWMHints())) { fprintf(stderr, "win_common_set_window_icon: error\n"); exit(1); }

  win_hints->flags = StateHint | IconPositionHint | IconPixmapHint | IconMaskHint;

  status = XpmCreatePixmapFromData (display, win, (char **)icon, &icon_pixmap, &mask_pixmap, NULL);

  if (status != 0) {
    fprintf(stderr, "win_common_set_colorWindowIcon error: %d\n", status); exit (1);
  }

  win_hints->initial_state = IconicState;
  win_hints->icon_pixmap = icon_pixmap;
  win_hints->icon_mask = mask_pixmap;
  win_hints->icon_x = 0; win_hints->icon_y = 0;

  XSetWMHints(display, win, win_hints);

  XFree(win_hints);
}



/* Матрица для заполнения градиентов */
const int tc_matrix [8] [3] =
  {
    {1, 1, 1} ,
    {1, 1, 0} ,
    {1, 0, 1} ,
    {0, 1, 1} ,
    {1, 0, 0} ,
    {0, 0, 1} ,
    {0, 1, 0} ,
  };
/* Создать набор градиентов */
void win_common_create_gradient (GC g, Pixmap * px)
{
  int i;
  int k;
  int step;
  XColor cl;

  step = 65535 / 200;

  for (k=0; k<7; k++) {
    for (i=0; i<201; i++) {
      win_common_set_color (&cl, tc_matrix [k] [0] * i * step,
		tc_matrix [k] [1] * i * step,
		tc_matrix [k] [2] * i * step);
      XSetForeground (display, g, cl.pixel);
      XDrawLine(display, *px, g, i, (k*10), i, ((k+1)*10));
    }
  }
}
