/* w_main.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>

#include "h_w_main.h"
#include "h_w_tool.h"
#include "h_w_command.h"
#include "h_w_common.h"
#include "h_com_console.h"
#include "h_c_text.h"
#include "h_LCP_reference_point.h"
#include "h_ico_color.h"
#include "h_o_cursor.h"
#include "h_coords.h"
#include "h_fs.h"



static char * m_wName = "LineColorPaint";
static long int m_event_mask;
static XColor m_clr; /* Цвет фона окна. */
static XColor tmp_clr; /* Цвет для отладочных сообщений в окне */
static char strBuff [NUMBUFFSIZE];
static unsigned int mBUTTON2;
static unsigned int time_parse;

static void win_main_init (void);
static void win_main_destruct (void);
static void win_main_mem_free (void);
static void win_main_key_press (KeySym keysym);
static void win_main_button_press (int x, int y, unsigned int button);
static void win_main_button_release (int x, int y, unsigned int button);
static void win_main_set_con_args2 (void);
static void win_main_set_con_args4 (void);
static void win_main_select_point (void);


void * win_main (void * unused)
{
  XEvent report;
  Atom pnprotocol [2];
  Atom wmprotocols;

  int depth;
  int buffSz = 20;
  char buff_kb [buffSz];
  KeySym keysym;
  XComposeStatus compose;

  if ((display = XOpenDisplay (NULL)) == NULL) {
    fprintf (stderr, "%s: cannot connect to X server\n", "w_main");
    exit (1);
  }
  screen = DefaultScreen (display);

  /* Устанавливаю графические данные */
  win_main_init ();

  /* Геометрические параметры дисплея и окна */
  display_width = DisplayWidth (display, screen);
  display_height = DisplayHeight (display, screen);

  /* Создать окно, установить цвет и прорисовать */
  m_window = XCreateWindow (display, RootWindow (display, screen),
			    0, 0, m_win_width, m_win_height, m_WIN_BORDER,
			    CopyFromParent, CopyFromParent,
			    CopyFromParent, 0, 0);
  XMapWindow (display, m_window);
  XMoveWindow (display, m_window, m_x_pos, m_y_pos);
  XFlush (display);
  XSelectInput (display, m_window, m_event_mask);
  XEventsQueued (display, QueuedAlready);

  /* Получение графического контекста */
  m_valuemask = GCFunction;
  m_values.function = GXcopy;
  depth = DefaultDepth (display, screen);
  m_pxm = XCreatePixmap (display, m_window, m_win_width, m_win_height, depth);
  if ((m_gc = XCreateGC (display, m_pxm, m_valuemask, &m_values)) < 0) {
    fprintf (stderr, "cannot create GC \n"); exit(1);
  }
  win_common_set_color (&m_clr, 0, 0, 0);
  XSetWindowBackground (display, m_window, m_clr.pixel);
  win_main_set_background ();  /* Установка фона для "pmx" */

  /* Установки менеджера окон */
  pnprotocol[0] = XInternAtom (display, "WM_TAKE_FOCUS", True);
  pnprotocol[1] = XInternAtom (display, "WM_DELETE_WINDOW",True);
  wmprotocols = XInternAtom(display, "WM_PROTOCOLS", True);
  XSetWMProtocols (display, m_window, pnprotocol, 2);

  /* Установить заголовок окна */
  XStoreName (display, m_window, m_wName);
  win_common_set_color_window_icon (m_window, LCP_color_xpm, icon_width, icon_height);

  
  /* Заполнить координаты, создать курсор */
  coords_init ();
  cursor_init ();

  /* Обновить окно */
  win_common_redraw_window (m_window);

  /* Панель инструментов */
  win_tool ();
  /* Графическая консоль */
  if (!XWndComCon) win_com ();

  /* Прочитать данные из файла */
  if (fptr) {
    fs_read_file (fptr);	  
    coords_zoom ();
    win_main_draw ();
  }
    
  /* handler events (Обработчик событий)*/
  while (1) {
    XNextEvent (display, &report);

    switch (report.type) {
    case Expose:
      if (report.xclient.window == t_window)
	win_tool_draw ();
      else 
	win_main_draw ();
      if (!XWndComCon) win_com_draw ();
      break;
    case ClientMessage:
      if (report.xclient.message_type == wmprotocols) {
	if (report.xclient.data.l [0] == pnprotocol [0]) {
	  ;
	} else if (report.xclient.data.l [0] == pnprotocol [1]) {
	  if (report.xclient.window == m_window) {
	    XCloseDisplay(display);
	    exit(0);
	  }
	}
      }
      break;
    case KeyPress:
      XLookupString (&report.xkey, buff_kb, buffSz, &keysym, &compose);
      if (report.xclient.window == m_window || report.xclient.window == c_window) {
	win_main_key_press (keysym);
	if (!XWndComCon) win_com_key_press (&report, buff_kb, keysym);
      } else if (report.xclient.window == t_window) {
	win_tool_key_press (keysym);
      } else {
	;
      }
      break;
    case KeyRelease:
      XLookupString (&report.xkey, buff_kb, buffSz, &keysym, &compose);
      if (report.xclient.window == m_window || report.xclient.window == c_window) {
	if (!XWndComCon) win_com_key_release (&report, buff_kb, keysym);
      } else if (report.xclient.window == t_window) {
	;
      } else {
	;
      }
      break;
    case ButtonPress:
      if (report.xclient.window == m_window) {
	win_main_button_press (report.xbutton.x, report.xbutton.y, report.xbutton.button);
      } else if (report.xclient.window == c_window) {
	win_com_button_press (report.xbutton.x, report.xbutton.y);
      } else if (report.xclient.window == t_window) {
	win_tool_button_press (report.xbutton.x, report.xbutton.y);
      } else {
	;
      }
      break;
    case ButtonRelease:
      if (report.xclient.window == m_window) {
	win_main_button_release (report.xbutton.x, report.xbutton.y, report.xbutton.button);
      }
      break;
    case MotionNotify:
      if (report.xclient.window == m_window) {
	CrCurs.X = report.xmotion.x;
	CrCurs.Y = report.xmotion.y;

	if (mBUTTON2)
	  coords_pan (); /* Panorama */

	win_common_redraw_window (m_window);
      }
      break;
    default:
      break;
    }

    XPeekEvent (display, &report);
  }

  win_main_destruct ();
  XCloseDisplay (display);
}

/* Установка данных */
static void win_main_init (void)
{
  time_parse = 0;
  step = (65535 / 256);
  m_valuemask = 0;
  m_event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask |
    ButtonReleaseMask | EnterWindowMask | LeaveWindowMask | PointerMotionMask
    | Button1MotionMask | Button2MotionMask | Button3MotionMask |
    Button4MotionMask | Button5MotionMask | ButtonMotionMask | KeymapStateMask
    | ExposureMask | VisibilityChangeMask | StructureNotifyMask |
    SubstructureNotifyMask | SubstructureRedirectMask | FocusChangeMask |
    PropertyChangeMask | ColormapChangeMask | OwnerGrabButtonMask | ResizeRedirectMask;
  win_common_set_color (&pt_clr, 65535, 65535, 65535); /* Общий цвет для всех точек. */
  mBUTTON2 = 0;
}

/* Освобождение памяти (для всех объектов) */
static void win_main_destruct (void)
{
  /* Tool window */
  win_tool_mem_free ();

  /* Command window */
  win_com_mem_free ();

  /* Main window */
  win_main_mem_free ();
}

#define INFO_CORNER 1
/* Основная функция вывода графики */
void win_main_draw (void)
{
# if INFO_CORNER /* corner for info  */
  win_main_change_mask (GXcopy);
  win_common_set_color (&tmp_clr, 10000, 10000, 10000);
  XSetForeground (display, m_gc, tmp_clr.pixel);
  XFillRectangle ( display, m_pxm, m_gc, 0, 0, 200, 50);

  win_common_set_color (&tmp_clr, 65535, 65535, 65535);
  XSetForeground (display, m_gc, tmp_clr.pixel);
  sprintf (strBuff, "X = %ld Y= %ld", rCrX, rCrY);
  XDrawString (display, m_pxm, m_gc, 10, 15, strBuff, strlen (strBuff));
  snprintf (strBuff, 200, "red=%li green=%li blue=%li", red, green, blue);
  XDrawString (display, m_pxm, m_gc, 10, 30, strBuff, strlen (strBuff));
# endif

  cursor_draw ();

  /* Окончательный вывод графики на экран */
  win_main_change_mask (GXcopy);
  XCopyArea ( display, m_pxm, m_window, m_gc, 0, 0,
	      m_win_width, m_win_height, 0, 0);
}

/* Обработчик событий клавиатуры */
static void win_main_key_press (KeySym keysym)
{

}

/* Обработчик событий мыши */
static void win_main_button_press (int x, int y, unsigned int button)
{
  if (button == Button2) {
    mBUTTON2 = 1;
    return;
  } else if (button == Button4) {
    coords_zoom_in ();
    return;
  }
#if 0
  else if (button == Button5) {
    coords_zoom_out ();
    return;
  }
#endif
  if (!IsParam) return;

  if (button == Button1) {
    switch (nCmd) {
    case DM_POINT:
      win_main_set_con_args2 (); /* Создать точку с указанными координатами. */
      break;
    case DM_POINT_DELETE:	/* Удалить точку с указанными координатами. */
      win_main_select_point ();
      break;
    case DM_LINE:  /* Создать линию с указанными координатами. */
      win_main_set_con_args4 ();
      break;
    case DM_LINE_DELETE: /* Удалить линию с указанными координатами. */
      win_main_set_con_args4 ();
      break;
    default:
      break;
    }
  }
}

/* Обработчик событий мыши */
static void win_main_button_release (int x, int y, unsigned int button)
{
  if (button == Button2) {
    mBUTTON2 = 0;
    coords_pan_after (); /* Panorama */
    win_common_redraw_window (m_window);
    return;
  }

  if (button == Button5) {
    coords_zoom_out ();
    return;
  }
}

/* Указать 4 точки и передать в командную строку */
static void win_main_set_con_args4 (void)
{
  static char tmpstr [TXT_MAX_STR];
  float X, Y;
  X = ((float)rCrX);
  Y = ((float)rCrY);

  X = X - zeroPt.x;
  Y = Y - zeroPt.y;
  X = X/m_scale;
  Y = Y/m_scale;

  switch (time_parse) {
  case 0:
    strcpy (con_str, "");
    strcpy (tmpstr, "");
    sprintf (con_str, "%.1f %.1f", X, Y);
    strcpy (tmpstr, con_str);
    time_parse++;
    break;
  case 1:
    sprintf (con_str, " %.1f %.1f", X, Y);
    if (((strlen (con_str)) + (strlen (tmpstr))) > TXT_MAX_STR) {
      time_parse = 0;
      return;
    }
    strcat (tmpstr, con_str);
    strcpy (con_str, tmpstr);
    time_parse = 0;
    break;
  default: break;
  }
  cursor_hide ();
  win_com_set_focus ();
  win_com_change_pos (strlen (con_str));
  win_com_refresh ();
  win_com_draw ();
}

/* Указать точку и передать в командную строку */
static void win_main_set_con_args2 (void)
{
  float X, Y;
  X = ((float)rCrX);
  Y = ((float)rCrY);

  fflush (stdout);
  X = X - zeroPt.x;
  Y = Y - zeroPt.y;
  X = X/m_scale;
  Y = Y/m_scale;

  sprintf (con_str, "%.5f %.5f", X, Y);
  
  cursor_hide ();
  win_com_set_focus ();
  win_com_change_pos (strlen (con_str));
  win_com_refresh ();
  win_com_draw ();
}

/* Выделить точку в окне, если точка находится в списке  */
static void win_main_select_point (void)
{
  pRefPointList nextPoint = NULL;

  float X, Y;
  X = ((float)rCrX);
  Y = ((float)rCrY);

  fflush (stdout);
  X = X - zeroPt.x;
  Y = Y - zeroPt.y;
  X = X/m_scale;
  Y = Y/m_scale;

  nextPoint = ref_point_lookup_next (X, Y);
  if (!nextPoint) return;
  X = nextPoint->pt.x;
  Y = nextPoint->pt.y;
  sprintf (con_str, "%.5f %.5f", X, Y);
  
  cursor_hide ();
  win_com_set_focus ();
  win_com_change_pos (strlen (con_str));
  win_com_refresh ();
  win_com_draw ();
}
  

/* Установка фона для "pmx" */
void win_main_set_background (void)
{
  win_main_change_mask (GXcopy);
  XSetForeground (display, m_gc, m_clr.pixel);
  XFillRectangle (display, m_pxm, m_gc, 0, 0, m_win_width, m_win_height);
}

/* Освободить память */
static void win_main_mem_free (void)
{
  XFreeGC (display, m_gc);
  XCloseDisplay (display);
}
