/* w_tool.c
 * Copyright (C) 2011, 2012 Zykov Anton
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "h_w_tool.h"
#include "h_w_main.h"
#include "h_w_common.h"
#include "h_c_tab.h"
#include "h_c_label.h"
#include "h_img_grad.h"
#include "h_ico_color.h"

#define t_WIN_BORDER 2
#define NUM_TEXT_LINES 20
#define LABEL_X 2
#define LABEL_WIDTH 196

char * documentation [NUM_TEXT_LINES] = {
  "v   (view)", "p   (rpoint)", "l   (line)", "pl  (pline)", "dp  (rpoint_delete)",
  "dl  (line_delete)", "dld (line_delete_dsc)", "dpl (pline_delete)", "-p  (rpoint_all_delete)", "-l  (line_all_delete)",
  "-pl (pline_all_delete)", "st  (show_tool)", "ut  (unshow_tool)", "of  (fs_open_file)", "cf  (fs_close_file)",
  "wf  (fs_write_file)", "", "", "", "",
};

#define win_tool_draw_line(x1, y1, x2, y2)				\
  XDrawLine(display, *pxm, t_gc, x1, (t_WIN_HEIGHT-y1), x2, (t_WIN_HEIGHT-y2))
#define win_tool_draw_rect(x, y, width, height) \
  XDrawRectangle(display, *pxm, t_gc, x, (t_WIN_HEIGHT-y-height), width, height)
#define win_tool_fill_rect(x, y, width, height) \
  XFillRectangle(display, *pxm, t_gc, x, (t_WIN_HEIGHT-y-height), width, height)
#define	win_tool_draw_str(x, y, str, len) \
  XDrawString(display, *pxm, t_gc, x, (t_WIN_HEIGHT-y), str, len)
#define win_tool_change_mask(f1) \
  win_common_change_mask(t_gc, t_valuemask, t_values, f1)
static void win_tool_init (void);
/* Пользовательские функции для заполнения массива "fTab" */
static void win_tool_draw_tab_0 (GC _g, Pixmap * px, int x, int y);
static void win_tool_draw_tab_1 (GC _g, Pixmap * px, int x, int y);
static void win_tool_draw_tab_2 (GC _g, Pixmap * px, int x, int y);

static Pixmap * pxm; /* Указатель на текущую вкладку */
static const int VSPACE = 1;
static LABEL lbl_color;
static char lbl_color_text [TXT_MAX_STR];
static LABEL lbl_comment1;
static RECT lbl_comment_rect1;
static char lbl_comment_text1 [TXT_MAX_STR] = "Console commands";
static LABEL lbl_comment2;
static RECT lbl_comment_rect2;
static char lbl_comment_text2 [TXT_MAX_STR] = " (long version):";
static int lbl_doc_height;
static LABEL lbl_doc [NUM_TEXT_LINES];
static RECT lbl_doc_rect [NUM_TEXT_LINES];
static char * t_wName = "Tools";
static unsigned long t_valuemask;
static long int t_event_mask;
static GC t_gc;
static XGCValues t_values;
static XImage * ximg; /* Изображение градиента */
static XImage * wimg; /* Копия градиента для получения цвета по текущим координитам */
static XFontStruct * t_foinf;

int win_tool (void)
{
  Atom pnprotocol[2];
  Atom wmprotocols;
  XColor clr_bk;

  win_tool_init ();  

  /* Создать окно, установить цвет и прорисовать */
  t_window = XCreateWindow (display, RootWindow (display, screen),
			    t_x_pos, t_y_pos, t_win_width, t_win_height, t_WIN_BORDER,
			    CopyFromParent, CopyFromParent,
			    CopyFromParent, 0, 0);
  XMapWindow (display, t_window);
  XMoveWindow (display, t_window, t_x_pos, t_y_pos);
  XFlush (display);
  XSelectInput (display, t_window, t_event_mask);
  XEventsQueued (display, QueuedAlready);
  
  /* Получение графического контекста */
  t_valuemask = GCFunction | GCLineWidth;
  t_values.line_width = 2;
  t_values.function = GXcopy;
  if ((t_gc = XCreateGC (display, t_window, t_valuemask, &t_values)) < 0) {
    fprintf(stderr, "cannot create GC \n"); exit(1);
  }
  win_common_set_color (&clr_bk, 58623, 58623, 58623);
  XSetWindowBackground (display, t_window, clr_bk.pixel);

  /* Установки менеджера окон */
  pnprotocol[0] = XInternAtom (display, "WM_TAKE_FOCUS", True);
  pnprotocol[1] = XInternAtom (display, "WM_DELETE_WINDOW",True);
  wmprotocols = XInternAtom (display, "WM_PROTOCOLS", True);
  XSetWMProtocols (display, t_window, pnprotocol, 2);

  /* Установить заголовок окна */
  XStoreName (display, t_window, t_wName);
  win_common_set_color_window_icon (t_window, LCP_color_xpm, icon_width, icon_height);

  /* Установить шрифт окна "t_font_name" */
  if (select_font) {
    if ((t_foinf = XLoadQueryFont (display, t_font_name))) { XSetFont (display, t_gc, t_foinf->fid); }
  }
  
  /* Создание вкладок */
  RECT t_rect = {0, 0, t_win_width, t_win_height};
  tab_init = win_tool_tab_init;
  tab_create (t_window, t_gc, t_rect, &pxm, t_valuemask, t_values);

  /* Обновить окно */
  win_tool_draw ();

  return (0);
}

static void win_tool_init (void)
{
  t_valuemask = 0;
  t_event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask |
  ButtonReleaseMask | EnterWindowMask | LeaveWindowMask | PointerMotionMask
  | Button1MotionMask | Button2MotionMask | Button3MotionMask |
  Button4MotionMask | Button5MotionMask | ButtonMotionMask | KeymapStateMask
  | ExposureMask | VisibilityChangeMask | SubstructureNotifyMask |
  SubstructureRedirectMask | FocusChangeMask | PropertyChangeMask |
  ColormapChangeMask | OwnerGrabButtonMask;
}

/* Основная функция вывода графики */
void win_tool_draw (void)
{
  win_tool_change_mask (GXcopy);
  tab_draw (t_window, t_gc, pxm);
}

/* Обработчик событий клавиатуры */
void win_tool_key_press (KeySym keysym)
{
  tab_key_press (&pxm, keysym);
  win_tool_draw ();
}

/* Обработчик событий мыши */
void win_tool_button_press (int x, int y)
{
  tab_button_press (t_gc, &pxm, x, y);
  win_tool_draw ();
}

/* Показать окно */
void win_tool_show (void)
{
  XMapWindow (display, t_window);
  XMoveWindow (display, t_window, t_x_pos, t_y_pos);
  win_common_redraw_window (t_window);
}

/* Скрыть окно */
void win_tool_unshow (void)
{
  XUnmapWindow (display, t_window);
  XMoveWindow (display, t_window, t_x_pos, t_y_pos);
  win_common_redraw_window (t_window);
}

#define CONT_HEIGHT 82
#define RND_VSPACE 10
/* Пользовательская функция.
   В ней необходио указать массив функций "fTAB[...]"
   Создание элементов управления.
*/
void win_tool_tab_init (GC g, Pixmap ** px)
{
  /* Установить по умолчанию функции обработки сообщений от вкладок */
  fTAB [0] = win_tool_draw_tab_0;
  fTAB [1] = win_tool_draw_tab_1;
  fTAB [2] = win_tool_draw_tab_2;

  /* Обновить окно и вкладки */
  *px = &tab[0]; /* Выбрать вкладку "0" и создать в ней градиент */
  if ((win_common_load_img_from_data (grad_xpm, &ximg)) == 0) /* Загрузка файла градиента */
    XPutImage (display, tab[0], g, ximg, 0, 0, 0, 0, 200, 70);
  else
    win_common_create_gradient (g, *px);

  tab_tabs_draw (g, *px); /* Нарисовать вкладки */
  *px = &tab[0]; /* Текущая вкладка "0" */

  /* Создание копии изображения вкладки 'tab[0]' для получения цвета палитры */
  wimg = XGetImage (display, tab[0], 0, 0, 200, 70, AllPlanes, XYPixmap);

  /* Создать сообщение */
  RECT rect = {.x=2, .y=CONT_HEIGHT, .width=196, .height=-1};  
  label_create (display, **px, screen, g, lbl_color_text, &rect, &lbl_color);

  int i;
  lbl_doc_height = label_get_font_height (g);
  int vspace = VSPACE + lbl_doc_height;
  int curr_height = vspace + CONT_HEIGHT + RND_VSPACE;

  lbl_comment_rect1.x=LABEL_X; lbl_comment_rect1.y=curr_height;
  lbl_comment_rect1.width=LABEL_WIDTH; lbl_comment_rect1.height=lbl_doc_height;
  lbl_comment1.flags = NO_FRAME;
  label_create (display, **px, screen, g, lbl_comment_text1, &lbl_comment_rect1, &lbl_comment1);
  curr_height += vspace;

  lbl_comment_rect2.x=LABEL_X; lbl_comment_rect2.y=curr_height;
  lbl_comment_rect2.width=LABEL_WIDTH; lbl_comment_rect2.height=lbl_doc_height;
  lbl_comment2.flags = NO_FRAME;
  label_create (display, **px, screen, g, lbl_comment_text2, &lbl_comment_rect2, &lbl_comment2);
  curr_height += vspace;
  
  for (i=0; i<NUM_TEXT_LINES; i++) {
    lbl_doc_rect [i].x=LABEL_X; lbl_doc_rect [i].y=curr_height;
    lbl_doc_rect [i].width=LABEL_WIDTH; lbl_doc_rect [i].height=lbl_doc_height;
    label_create (display, **px, screen, g, documentation [i], &lbl_doc_rect [i], &lbl_doc [i]);
    curr_height += vspace;
  }
  
  /* Установить произвольный цвет из палитры */
  fTAB[0] (g, *px, 30, 30);
}

/* Пользовательская функция (Вкладка 0).
   Обработать нажатие клавиши мыши в точке (x, y). */
void win_tool_draw_tab_0 (GC _g, Pixmap * px, int x, int y)
{
  static unsigned long r, g, b;
  int coordY = t_WIN_HEIGHT - y;

  /* Если выбрана палитра, то получить цвет палитры в данной точке. */
  if ((coordY > 630) && (coordY < 700)) {
    char lbl_str [TXT_MAX_STR];
    int x1=0, y1=620, width1=200, height1=10;
    int x2=2, y2=622, width2=196, height2=6;

    r = g = b = 0;

    tPixval = XGetPixel (wimg, x, y);
    win_common_get_RGB_val (tPixval, &r, &g, &b);
    snprintf (lbl_str, 200, "red=%li green=%li blue=%li", r, g, b);

    red = r * step;
    green = g * step;
    blue = b * step;

    win_common_set_color (&clr, 35000, 35000, 35000); XSetForeground (display, _g, clr.pixel);
    XFillRectangle(display, *px, _g, x1, (t_WIN_HEIGHT-y1-height1), width1, height1);
    win_common_set_color (&clr, red, green, blue); XSetForeground (display, _g, clr.pixel);
    XFillRectangle(display, *px, _g, x2, (t_WIN_HEIGHT-y2-height2), width2, height2);

    strcpy (lbl_color.text, lbl_str);
    label_draw (display, *px, _g, &lbl_color);

    label_draw (display, *px, _g, &lbl_comment1);
    label_draw (display, *px, _g, &lbl_comment2);    
    
    int i;
    for (i=0; i<NUM_TEXT_LINES; i++) {
      label_draw (display, *px, _g, &lbl_doc [i]);
    }
  }
}

/* Пользовательская функция (Вкладка 1). */
void win_tool_draw_tab_1 (GC _g, Pixmap * px, int x, int y)
{

}

/* Пользовательская функция (Вкладка 2). */
void win_tool_draw_tab_2 (GC _g, Pixmap * px, int x, int y)
{

}

/* Освободить память */
void win_tool_mem_free (void)
{
  XDestroyImage (ximg);
  XDestroyImage (wimg);
  tab_mem_free ();
  XFreeFont (display, t_foinf);
  XFreeGC (display, t_gc);
  free (t_font_name);
}
